#neutrino - framework for distributed web-service development.[![Build Status](https://secure.travis-ci.org/DenisRechkunov/neutrino.png)](http://travis-ci.org/DenisRechkunov/neutrino)

##Description

The main aim of the project is to solve a scaling issue for node.js web-services. Node.js has an implementation of cluster but it's only locally on one physical machine.

Neutrino gives ability to create web-service which is distributed geographically on different physical or virtual servers around the world. Also neutrino can automatically balance your cluster and synchronize state of its worker nodes it also uses safe protocol for communication using key pair and TLS (optional). As a main data storage it uses MongoDB which also have a good scaling features. 

Neutrino isn't just a framework for server but client-side code too. It uses Socket.IO as a data channel between client and server script and provides with high-level engine for neutrino usage.

Neutrino has MVC pattern implementation with observable properties in model, auto validators in controller and subscription engine on client to receive model's changes in real time.

If you need to send data to whole distributed neutrino system you can create "event service". It will send data to master node which will send it to most free node for data processing.

##Installation

To install neutrino package just use NPM:

```bash
npm install neutrino
```

Or download last release in tags tab of this page.


##Examples

To start neutrino system you must start master and at least one worker node.

```javascript
var neutrino = require('neutrino');

neutrino.configure({
    ... // config options
});

neutrino.createMaster();
neutrino.createWorker();
```

If you want to start more workers you need change worker node's host and/or port configuration. See [default config](https://github.com/DenisRechkunov/neutrino/blob/master/defaults.json) for details.
All config options were passed to "configure" method will override default config options, but you don't need to set all config options here, just options which differs from defaults.

###Model example

```javascript
module.exports = TestModel;

function TestModel(config) {

    // call the ModelBase constructor
    TestModel.super_.call(this, config);

    this.on('dataFromService', function (sender, data) {
        // here processing data from event service with name equal sender
    });

    var self = this;
    // here we change a property every second for testing change event
    setInterval(function(){
        self.testProperty1.$((new Date()).getTime());
    },1000);
}

// properties
TestModel.prototype.testProperty1 = 'testValue1';
TestModel.prototype.testProperty2 = 'testValue2';
TestModel.prototype._testPrivateProperty1 = 'testPrivateValue1';
TestModel.prototype.testPrivateProperty2_ = 'testPrivateValue2';

// methods
TestModel.prototype.testMethod = function (someArg1, someArg2, someArg3, callback) {
    //processing args and invoke callback with result
    callback(someArg1 + someArg2 + someArg3);
};

TestModel.prototype.testPrivateMethod_ = function (someArg1, someArg2, someArg3, callback) {
    callback(someArg1 + someArg2 + someArg3);
};

TestModel.prototype.sendToEventServiceExample_ = function (serviceName, dataObject) {
    this.emit('sentToService', serviceName, dataObject);
};

TestModel.prototype.sendNotificationToUser_ = function (sessionId, dataObject) {
    this.emit('notify', sessionId, dataObject);
};
```

###Controller example

```javascript
module.exports = TestController;

function TestController(config, model, view) {
    // call base controller constructor
    TestController.super_.call(this, config, model, view);
}

TestController.prototype.accessValidator = function (sessionId, callback) {
    // check user by sessionId for whole model access and invoke callback
    callback(new Error('Access denied'));
};

TestController.prototype.testProperty1GetValidtor = function(sessionId, callback){
    // check user by sessionId for model property get access
    callback(null);
};

TestController.prototype.testProperty1SetValidtor = function(newValue, sessionId, callback){
    // check user by sessionId for model property set access and new value validation
    callback(null);
};

TestController.prototype.testMethodInvokeValidator = function(sessionId, callback,
    someArg1, someArg2, someArg3){
    // check user by sessionId for model method invoke access and invocation arguments
    callback(new Error('Wrong arguments'));
};
```

###Event service example

```javascript
module.exports = TestService;

function TestService(config, sendDataHandler) {

    var self = this;
    setInterval(function () {

        sendDataHandler('TestModel', {
            messageText:'testMessage'
        });
        // this message object will be received by model with constructor name 'TestModel'

        sendDataHandler(null, {
            messageText:'testMessage2'
        });
        // this message object will be received by all models

    }, 5000);
}

TestService.prototype.handleData = function (modelName, data) {
    //processing data from any model
};
```

###Client-side usage example

index.html
```html
<!DOCTYPE html>
<html>
<head>
    <title>Neutrino client example</title>
    <!-- Use a path to neutrino-client folder here -->
    <script data-main="app.js" src="libs/require.js"></script>
</head>
<body>

</body>
</html>
```

app.js
```javascript
require(['./neutrino/mvc/view-hub-client.js'], function (ViewHubClient) {

    var model, client;

    client = new ViewHubClient('http://localhost:8080');

    client.on('connect', function (address) {
        alert('connected: ' + address.host + ':' + address.port);
    });

    client.on('disconnect', function () {
        alert('disconnected');
    });

    client.on('reconnect', function () {
        alert('reconnected');
    });

    // listen to notification channel
    client.on('notification', function (viewName, notificationObject) {
        alert('received notification from ' + viewName);
    });

    model = client.getModelWrapper('TestModel');

    // catch all server-side errors
    model.$events.on('error', function (error) {
        alert(error);
    });

    model.$events.on('notification', function (obj) {
        alert(obj);
    });

    model.$events.on('ready', function () {

        // call model method
        model.testMethod('Hello', 'Great', 'World', function (error, result) {
            if (error) {
                throw error;
            }
            console.log(result);
        });

        // get current property value
        console.log('current value' + model.testProperty1.$());

        // subscribe to property changed event
        model.$events.on('testProperty1Changed', function (oldValue, newValue) {
            console.log(oldValue + '->' + newValue);
        });

        // set value to property
        model.testProperty1.$('hello world');

    });
});
```

##Project development

To be a contributor of this project, please read [Google JavaScript Style Guide](http://google-styleguide.googlecode.com/svn/trunk/javascriptguide.xml) and follow the guide instructions before commit or pull request. To document new code please use [JsDoc toolkit](http://code.google.com/p/jsdoc-toolkit/) and its [tag reference](http://code.google.com/p/jsdoc-toolkit/wiki/TagReference).

Denis Rechkunov <denis.rechkunov@gmail.com>