/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var net = require('net'),
	util = require('util'),
	tls = require('tls'),
	events = require('events'),
	Scheduler = require('../../common/patterns/scheduler.js'),
	eventBus = require('./common.js');

module.exports = EventBusClient;

util.inherits(EventBusClient, events.EventEmitter);

/**
 * Create new instance of neutrino event bus client.
 * @param {Config} config Neutrino config object.
 * @constructor
 * @extends events.EventEmitter
 */
function EventBusClient(config) {

	events.EventEmitter.call(this);

	this.config_ = config;

	// bind handlers
	this.connectionHandler_ = this.connectionHandler_.bind(this);
	this.closeHandler_ = this.closeHandler_.bind(this);
	this.incomingMessageHandler_ = this.incomingMessageHandler_.bind(this);
	this.sendMessageQueueHandler_ = this.sendMessageQueueHandler_.bind(this);
	this.socketDataHandler_ = this.socketDataHandler_.bind(this);
	this.messageWriteToLogHandler_ = this.messageWriteToLogHandler_.bind(this);
	this.connect = this.connect.bind(this);

	// save often-used parameters
	this.masterSecret_ = this.config_.$(EventBusClient.parameters.masterSecret);
	this.workerSecret_ = this.config_.$(EventBusClient.parameters.workerSecret);
	this.sendMessagePerTimeLimit_ = this.config_.$(EventBusClient.parameters.sendMessagePerTimeLimit);

	// init
	this.offlineMessageQueue_ = [];
	this.onlineMessageQueue_ = [];
	this.serverAddressString_ = util.format('%s:%d',
		this.config_.$(EventBusClient.parameters.serverAddress),
		this.config_.$(EventBusClient.parameters.serverPort));

	this.reconnectScheduler_ = new Scheduler(
		this.config_.$(EventBusClient.parameters.reconnectInterval),
		this.config_.$(EventBusClient.parameters.reconnectInterval),
		this.connect);

	this.sendMessageQueueScheduler_ = new Scheduler(0,
		this.config_.$(EventBusClient.parameters.sendMessageQueueInterval),
		this.sendMessageQueueHandler_);

	this.tlsConfig_ = eventBus.createTlsConfig(this.config_);

}

/**
 * Enum of event bus client events.
 * @enum {String}
 */
EventBusClient.events = {
	serviceMessage: 'serviceMessage',
	workerConnected: 'workerConnected',
	workerDisconnected: 'workerDisconnected',
	messageFromMaster: 'messageFromMaster'
};

/**
 * Set of config parameters for module.
 * @type {Object}
 */
EventBusClient.parameters = eventBus.parameters;

/**
 * Current config instance.
 * @type {Config}
 * @private
 */
EventBusClient.prototype.config_ = null;

/**
 * Current TLS config instance.
 * @type {Object}
 * @private
 */
EventBusClient.prototype.tlsConfig_ = null;

/**
 * Current master secret phrase.
 * @type {String}
 * @private
 */
EventBusClient.prototype.masterSecret_ = '';

/**
 * Current worker secret phrase.
 * @type {String}
 * @private
 */
EventBusClient.prototype.workerSecret_ = '';

/**
 * Current limit for message passing count per time.
 * @type {Number}
 * @private
 */
EventBusClient.prototype.sendMessagePerTimeLimit_ = 0;

/**
 * Message queue for delayed offline message sending.
 * @type {Array}
 * @private
 */
EventBusClient.prototype.offlineMessageQueue_ = null;

/**
 * Message queue for delayed online message sending.
 * @type {Array}
 * @private
 */
EventBusClient.prototype.onlineMessageQueue_ = null;

/**
 * EBS address and port string.
 * @type {String}
 * @private
 */
EventBusClient.prototype.serverAddressString_ = '';

/**
 * Reconnect attempt scheduler.
 * @type {Scheduler}
 * @private
 */
EventBusClient.prototype.reconnectScheduler_ = null;

/**
 * Send message queue scheduler.
 * @type {Scheduler}
 * @private
 */
EventBusClient.prototype.sendMessageQueueScheduler_ = null;

/**
 * Current socket state flag.
 * @type {Boolean}
 * @private
 */
EventBusClient.prototype.isConnected_ = false;

//noinspection JSValidateJSDoc
/**
 * TCP socket object.
 * @type {net.Socket}
 * @private
 */
EventBusClient.prototype.socketNamespace_ = null;

/**
 * Current socket address.
 * @type {Object}
 * @private
 */
EventBusClient.prototype.socketAddress_ = null;

/**
 * Current socket address and port string.
 * @type {String}
 * @private
 */
EventBusClient.prototype.socketAddressString_ = '';

/**
 * Connect to EBS.
 */
EventBusClient.prototype.connect = function () {

	if (this.config_.$(EventBusClient.parameters.tlsEnabled)) {

		this.socketNamespace_ = tls.connect(
			this.config_.$(EventBusClient.parameters.serverPort),
			this.config_.$(EventBusClient.parameters.serverAddress),
			this.tlsConfig_,
			this.connectionHandler_);

		this.socketNamespace_.socket.on('close', this.closeHandler_);

	} else {

		this.socketNamespace_ = net.connect(
			this.config_.$(EventBusClient.parameters.serverPort),
			this.config_.$(EventBusClient.parameters.serverAddress),
			this.connectionHandler_);

		this.socketNamespace_.on('close', this.closeHandler_);

	}

	this.socketNamespace_.setEncoding(this.config_.$(EventBusClient.parameters.charset));
	this.socketNamespace_.on('data', this.socketDataHandler_);
};

/**
 * Handles all data from socket.
 * @param {String} data Data representation.
 * @private
 */
EventBusClient.prototype.socketDataHandler_ = function (data) {
	var items = data.split('\r\n');
	items.forEach(this.incomingMessageHandler_);
};

/**
 * Handle EBS connection.
 * @private
 */
EventBusClient.prototype.connectionHandler_ = function () {

	this.socketAddress_ = this.socketNamespace_.address();
	this.socketAddressString_ = util.format('%s:%d',
		this.socketAddress_.address,
		this.socketAddress_.port);

	this.isConnected_ = true;
	this.reconnectScheduler_.stop();

	this.emit(EventBusClient.events.serviceMessage, {
		connection: util.format('%s->%s', this.socketAddressString_, this.serverAddressString_),
		message: this.tlsConfig_.enabled ? 'TLS/SSL connection established' : 'Connection established'
	});

	this.emit(EventBusClient.events.workerConnected);
	this.sendMessageQueueScheduler_.start();
};

/**
 * Handle socket close event.
 * @private
 */
EventBusClient.prototype.closeHandler_ = function () {

	this.isConnected_ = false;

	this.sendMessageQueueScheduler_.stop();
	this.reconnectScheduler_.start();
	this.onlineMessageQueue_ = [];

	this.emit(EventBusClient.events.workerDisconnected);

	this.emit(EventBusClient.events.serviceMessage, {
		connection: util.format('%s->%s', this.socketAddressString_, this.serverAddressString_),
		message: 'Connection lost'
	});
};

/**
 * Handle new incoming message from master.
 * @param {String} message Message JSON from master.
 * @private
 */
EventBusClient.prototype.incomingMessageHandler_ = function (message) {

	if (message.length === 0) {
		return;
	}

	try {

		var messageObject = typeof (message) === 'string' ? JSON.parse(message) : message;

		if (util.isArray(messageObject)) {
			messageObject.forEach(this.incomingMessageHandler_);
			return;
		}

		if (!messageObject.secret || messageObject.secret !== this.masterSecret_) {
			//noinspection ExceptionCaughtLocallyJS
			throw new Error('Wrong secret from master');
		}

	} catch (e) {
		this.socketNamespace_.destroy();
		throw e;
	}

	this.emit(EventBusClient.events.messageFromMaster, messageObject);
	this.emit(EventBusClient.events.serviceMessage, {
		connection: util.format('%s->%s', this.serverAddressString_, this.socketAddressString_),
		message: util.format('Master sent message, type - %s', messageObject.type)
	});
};

/**
 * Handles message and log it.
 * @param {Object} messageObject Message object.
 * @private
 */
EventBusClient.prototype.messageWriteToLogHandler_ = function (messageObject) {
	this.emit(EventBusClient.events.serviceMessage, {
		connection: util.format('%s->%s', this.socketAddressString_, this.serverAddressString_),
		message: util.format('Worker sent message, type - %s', messageObject.type)
	});
};

/**
 * Send message to Master.
 * @param {Object} messageObject Message object for Master.
 */
EventBusClient.prototype.sendToMaster = function (messageObject) {

	messageObject.secret = this.workerSecret_;

	switch (messageObject.type) {
		case eventBus.messageTypes.data:
		case eventBus.messageTypes.shared:
			this.offlineMessageQueue_.push(messageObject);
			break;
		default:
			if (!this.isConnected_) {
				return;
			}
			this.onlineMessageQueue_.push(messageObject);
	}
};

/**
 * Tries to send all messages from message queue.
 */
EventBusClient.prototype.sendMessageQueueHandler_ = function () {

	if (!this.isConnected_ ||
		(this.offlineMessageQueue_.length === 0 &&
			this.onlineMessageQueue_.length === 0)) {
		return;
	}

	var offlineMessageToSend = this.offlineMessageQueue_.splice(0, this.sendMessagePerTimeLimit_),
		onlineMessageToSend = this.onlineMessageQueue_.splice(0, this.sendMessagePerTimeLimit_),
		allMessagesToSend = offlineMessageToSend.concat(onlineMessageToSend),
		messages = JSON.stringify(allMessagesToSend);

	try {
		this.socketNamespace_.write(messages + '\r\n');
		allMessagesToSend.forEach(this.messageWriteToLogHandler_);
	} catch (e) {
		this.offlineMessageQueue_ = offlineMessageToSend.concat(this.offlineMessageQueue_);
		this.onlineMessageQueue_ = onlineMessageToSend.concat(this.onlineMessageQueue_);
	}
};