/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var net = require('net'),
	path = require('path'),
	tls = require('tls'),
	events = require('events'),
	util = require('util'),
	eventBus = require('./common.js');

module.exports = EventBusServer;

util.inherits(EventBusServer, events.EventEmitter);

/**
 * Create new instance of the Event Bus Server.
 * @param {Config} config Neutrino config object.
 * @constructor
 */
function EventBusServer(config) {

	events.EventEmitter.call(this);

	this.config_ = config;

	// bind handlers
	this.incomingConnectionHandler_ = this.incomingConnectionHandler_.bind(this);
	this.messageFromMasterHandler_ = this.messageFromMasterHandler_.bind(this);
	this.messageFromWorkerHandler_ = this.messageFromWorkerHandler_.bind(this);
	this.serverStartedHandler_ = this.serverStartedHandler_.bind(this);

	// save often used config parameters
	this.masterSecret_ = this.config_.$(EventBusServer.parameters.masterSecret);
	this.workerSecret_ = this.config_.$(EventBusServer.parameters.workerSecret);
	this.charset_ = this.config_.$(EventBusServer.parameters.charset);

	this.sockets_ = [];
	this.workersByAddress_ = {};
	this.workersByWorkerId_ = {};

	this.tlsConfig_ = eventBus.createTlsConfig(this.config_);

	this.server_ = this.config_.$(EventBusServer.parameters.tlsEnabled) ?
		tls.createServer(this.tlsConfig_, this.incomingConnectionHandler_) :
		net.createServer(this.incomingConnectionHandler_);

	this.on(EventBusServer.events.messageFromMaster, this.messageFromMasterHandler_);
}

/**
 * Enum of event bus server events.
 * @enum {String}
 */
EventBusServer.events = {
	serviceMessage: 'serviceMessage',
	messageFromMaster: 'messageFromMaster',
	messageFromWorker: 'messageFromWorker',
	workerConnected: 'workerConnected',
	workerDisconnected: 'workerDisconnected'
};

/**
 * Set of config parameters for module.
 * @type {Object}
 */
EventBusServer.parameters = eventBus.parameters;

/**
 * Current instance of config object.
 * @type {Config}
 * @private
 */
EventBusServer.prototype.config_ = null;

/**
 * Secret token for master of event bus.
 * @type {String}
 * @private
 */
EventBusServer.prototype.masterSecret_ = '';

/**
 * Current charset.
 * @type {String}
 * @private
 */
EventBusServer.prototype.charset_ = '';

/**
 * Secret token for worker of event bus.
 * @type {String}
 * @private
 */
EventBusServer.prototype.workerSecret_ = '';

/**
 * TLS configuration object.
 * @type {Object}
 * @private
 */
EventBusServer.prototype.tlsConfig_ = null;

/**
 * Array of connected sockets.
 * @type {Array}
 * @private
 */
EventBusServer.prototype.sockets_ = null;

/**
 * Hash table of workers by worker id.
 * @type {Object}
 * @private
 */
EventBusServer.prototype.workersByWorkerId_ = null;

/**
 * Hash table of workers by socket address.
 * @type {Object}
 * @private
 */
EventBusServer.prototype.workersByAddress_ = null;

//noinspection JSValidateJSDoc
/**
 * TCP server which EBS use to connect with EBCs;
 * @type {net.Server}
 * @private
 */
EventBusServer.prototype.server_ = null;

/**
 * Server address string.
 * @type {String}
 * @private
 */
EventBusServer.prototype.serverAddressString_ = '';

/**
 * Start to listen for incoming connections.
 */
EventBusServer.prototype.start = function () {
	this.server_.listen(this.config_.$(EventBusServer.parameters.serverPort), this.serverStartedHandler_);
};

/**
 * Handles started server event and log information.
 * @private
 */
EventBusServer.prototype.serverStartedHandler_ = function () {
	var serverAddress = this.server_.address();
	this.serverAddressString_ = util.format('%s:%d', serverAddress.address, serverAddress.port);
	this.emit(EventBusServer.events.serviceMessage, {
		connection: this.serverAddressString_,
		message: this.tlsConfig_.enabled ? 'EBS started with TLS/SSL' : ' EBS started'
	});
};

//noinspection JSUnusedGlobalSymbols
/**
 * Send event message to all Workers of the cluster or to specified worker.
 * @param {object} messageObject Event message.
 * @param {string} workerId Worker ID (Optional).
 */
EventBusServer.prototype.sendToWorkerById = function (messageObject, workerId) {
	this.emit(EventBusServer.events.messageFromMaster, messageObject, workerId);
};

/**
 * Send event message to worker
 * @param {Object} messageObject Message object.
 * @param {String} address Address of worker node.
 */
EventBusServer.prototype.sendToWorkerByAddress = function (messageObject, address) {

	var workerId = this.workersByAddress_[address].workerId;

	if (!workerId) {
		return;
	}

	this.emit(EventBusServer.events.messageFromMaster, messageObject, workerId);
};

/**
 * Sends message to all workers except specified.
 * @param {Object} messageObject Message object to send.
 * @param {String} exceptWorkerId Worker ID, which will not receive a message.
 */
EventBusServer.prototype.sendToWorkersExcept = function (messageObject, exceptWorkerId) {

	for (var workerId in this.workersByWorkerId_) {
		if (!this.workersByWorkerId_.hasOwnProperty(workerId) ||
			workerId === exceptWorkerId) {
			continue;
		}
		this.emit(EventBusServer.events.messageFromMaster, messageObject, workerId);
	}
};

//noinspection JSValidateJSDoc
/**
 * Process incoming connection.
 * @param {net.Socket} socket Incoming connection socket.
 * @private
 */
EventBusServer.prototype.incomingConnectionHandler_ = function (socket) {

	var self = this,
		socketAddress = this.getSocketAddress_(socket);

	socket.setEncoding(this.charset_);

	socket.on('close', function () {
		self.removeSocket_(socket, socketAddress);
		self.emit(EventBusServer.events.serviceMessage, {
			connection: util.format('%s->%s', socketAddress, self.serverAddressString_),
			message: 'Worker disconnected'
		});
		self.emit(EventBusServer.events.workerDisconnected, socketAddress);
	});

	socket.on('data', function (data) {
		var items = data.split('\r\n');
		items.forEach(function (item) {
			self.messageFromWorkerHandler_(socket, item);
		});
	});

	this.sockets_.push(socket);

	this.workersByAddress_[socketAddress] = {socket: socket, isAuthorized: false};

	this.emit(EventBusServer.events.serviceMessage, {
		connection: util.format('%s->%s', socketAddress, this.serverAddressString_),
		message: 'Worker connected'
	});
	this.emit(EventBusServer.events.workerConnected, socketAddress);
};

//noinspection JSValidateJSDoc
/**
 * Process message from worker and send to master.
 * @param {net.Socket} socket Connected socket.
 * @param {String} message JSON message string.
 * @private
 */
EventBusServer.prototype.messageFromWorkerHandler_ = function (socket, message) {

	if (message.length === 0) {
		return;
	}

	var messageObject,
		socketAddress = this.getSocketAddress_(socket),
		handler = this.messageFromWorkerHandler_,
		workerId;

	try {
		messageObject = typeof (message) === 'string' ?
			JSON.parse(message) :
			message;

		if (util.isArray(messageObject)) {

			messageObject.forEach(function (item) {
				handler(socket, item);
			});
			return;
		}

		if (!messageObject.secret || messageObject.secret !== this.workerSecret_) {
			//noinspection ExceptionCaughtLocallyJS
			throw new Error('Wrong secret from ' + socketAddress);
		}

		if (!messageObject.sender) {
			//noinspection ExceptionCaughtLocallyJS
			throw new Error('Worker must have ID');
		}

	} catch (e) {
		socket.destroy();
		throw e;
	}

	workerId = messageObject.sender;

	this.workersByAddress_[socketAddress].isAuthorized = true;
	this.workersByAddress_[socketAddress].workerId = workerId;
	this.workersByWorkerId_[workerId] = this.workersByAddress_[socketAddress];

	this.emit(EventBusServer.events.serviceMessage, {
		connection: util.format('%s:%d->%s',
			socket.remoteAddress,
			socket.remotePort,
			this.serverAddressString_),
		message: util.format('Worker sent message, type - %s', messageObject.type)
	});
	this.emit(EventBusServer.events.messageFromWorker, messageObject, workerId, socketAddress);
};

/**
 * Process message from Master and send it to worker(s).
 * @param {Object} messageObject Message object to send.
 * @param {String} workerId Id of worker (optional).
 * @private
 */
EventBusServer.prototype.messageFromMasterHandler_ = function (messageObject, workerId) {

	var masterSecret = this.masterSecret_,
		message,
		socket,
		socketAddress;

	if (util.isArray(messageObject)) {
		messageObject.forEach(function (item) {
			item.secret = masterSecret;
		});
	} else {
		messageObject.secret = masterSecret;
	}

	message = JSON.stringify(messageObject);

	if (workerId) {

		if (!(workerId in this.workersByWorkerId_) || !this.workersByWorkerId_[workerId].isAuthorized) {
			return;
		}

		socket = this.workersByWorkerId_[workerId].socket;
		socketAddress = socket.address();
		socket.write(message + '\r\n');

		this.emit(EventBusServer.events.serviceMessage, {
			connection: util.format('%s->%s:%d',
				this.serverAddressString_,
				socketAddress.address,
				socketAddress.port),
			message: util.format('Master sent message, type - %s', messageObject.type)
		});

	} else {

		this.sockets_.forEach(function (socket) {
			var currentWorkerAddress = this.getSocketAddress_(socket);
			if (!(currentWorkerAddress in this.workersByAddress_) ||
				!this.workersByAddress_[currentWorkerAddress].isAuthorized) {
				return;
			}
			socket.write(message + '\r\n');
		});
		this.emit(EventBusServer.events.serviceMessage, {
			connection: this.serverAddressString_,
			message: util.format('Master sent broadcast message, type - %s', messageObject.type)
		});
	}
};

//noinspection JSValidateJSDoc
/**
 * Gets socket address string.
 * @param {net.Socket} socket Worker socket object.
 * @return {String}
 * @private
 */
EventBusServer.prototype.getSocketAddress_ = function (socket) {
	return util.format('%s:%d', socket.remoteAddress, socket.remotePort);
};

//noinspection JSValidateJSDoc
/**
 * Remove socket from EBS list.
 * @param {net.Socket} socket Connected socket.
 * @param {String} socketAddress Address of socket was closed.
 * @private
 */
EventBusServer.prototype.removeSocket_ = function (socket, socketAddress) {

	var socketIndex = this.sockets_.indexOf(socket),
		workerId = this.workersByAddress_[socketAddress].workerId;

	this.sockets_.splice(socketIndex, 1);

	if (workerId) {
		delete this.workersByWorkerId_[workerId];
	}

	delete this.workersByAddress_[socketAddress];
};