/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var fs = require('fs'),
	path = require('path'),
	util = require('util'),
	Logger = require('../../security/logger.js'),
	logger;

module.exports = EventServiceProvider;

/**
 * Creates new instance of event service provider.
 * @param {Object} config System config object.
 * @param {String} eventServiceFolder Path to folder with event services.
 * @param {function} messageHandler Message handler for event service.
 * @constructor
 */
function EventServiceProvider(config, eventServiceFolder, messageHandler) {

	this.eventServicesFolder_ = eventServiceFolder;
	this.config_ = config;
	logger = new Logger(config);

	if (typeof (messageHandler) === 'function') {
		this.messageHandler_ = messageHandler;
	}

	this.initEventServices_();
}

/**
 * Current system config object.
 * @type {Object}
 * @private
 */
EventServiceProvider.config_ = null;

//noinspection JSUnusedLocalSymbols
/**
 * Current event service message handler.
 * @param {String} serviceName Name of event service.
 * @param {String} modelName Receiver model name.
 * @param {Object} data Message.
 */
EventServiceProvider.prototype.messageHandler_ = function (serviceName, modelName, data) {
	// nothing to do
};

/**
 * Current event services folder.
 * @type {String}
 * @private
 */
EventServiceProvider.prototype.eventServicesFolder_ = '';

/**
 * Current event services hash table.
 * @type {Object}
 * @private
 */
EventServiceProvider.prototype.eventServices_ = null;

/**
 * Initializes all events services of master node.
 * @private
 */
EventServiceProvider.prototype.initEventServices_ = function () {

	var self = this;
	self.eventServices_ = {};

	fs.readdir(self.eventServicesFolder_, function (error, files) {

		if (error) {
			return;
		}

		files.forEach(function (file) {

			if (path.extname(file) !== '.js') {
				return;
			}

			var servicePath = path.resolve(util.format('%s/%s', self.eventServicesFolder_, file)),
				serviceName = path.basename(file, '.js'),
				Service = require(servicePath),
				service = new Service(self.config_, function (modelName, data) {
					self.messageHandler_(serviceName, modelName, data);
				});

			logger.trace(util.format('Event service "%s" loaded', servicePath));

			self.eventServices_[serviceName] = service;

		});
	});

};

/**
 * Gets event services collection.
 * @return {Object}
 */
EventServiceProvider.prototype.getServices = function () {
	return this.eventServices_;
};