/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var util = require('util'),
	path = require('path'),
	url = require('url'),
	Logger = require('../../security/logger.js'),
	logger,
	eventBus = require('../event-bus/common.js'),
	EventBusServer = require('../event-bus/event-bus-server.js'),
	Balancer = require('./balancer.js'),
	EventServiceProvider = require('./event-service-provider.js'),
	NonCacheableWebServer = require('../../common/non-cacheable-web-server.js'),
	Scheduler = require('../../common/patterns/scheduler.js');

module.exports = MasterNode;

/**
 * Creates new instance of cluster master.
 * @param {Config} config Neutrino config object.
 * @constructor
 */
function MasterNode(config) {

	// bind handlers
	this.sendEventServiceMessageQueueHandler_ = this.sendEventServiceMessageQueueHandler_.bind(this);
	this.httpRequestHandler_ = this.httpRequestHandler_.bind(this);
	this.eventServiceMessageHandler_ = this.eventServiceMessageHandler_.bind(this);
	this.messageHandler_ = this.messageHandler_.bind(this);
	this.serviceMessageHandler_ = this.serviceMessageHandler_.bind(this);
	this.workerConnectHandler_ = this.workerConnectHandler_.bind(this);
	this.workerDisconnectHandler_ = this.workerDisconnectHandler_.bind(this);

	this.config_ = config;
	logger = new Logger(config);

	this.sendMessagePerTimeLimit_ = this.config_.$(MasterNode.parameters.sendMessagePerTimeLimit);

	// event service message queue
	this.eventServiceMessageQueue_ = [];
	this.eventServiceMessageSendScheduler_ =
		new Scheduler(0,
			this.config_.$(MasterNode.parameters.sendMessageQueueInterval),
			this.sendEventServiceMessageQueueHandler_);
	this.eventServiceMessageSendScheduler_.start();

	this.eventBusServer_ = new EventBusServer(config);
	this.balancer_ = new Balancer();

	this.webServer_ = new NonCacheableWebServer(this.config_.$(MasterNode.parameters.httpPort));
	this.webServer_.on(this.webServer_.events.onRequest, this.httpRequestHandler_);

	this.eventServiceProvider_ = new EventServiceProvider(
		config,
		config.$(MasterNode.parameters.eventServicesFolder),
		this.eventServiceMessageHandler_);

	this.workersByAddress_ = {};

	var eventBusServerEvents = EventBusServer.events;

	this.eventBusServer_.on(eventBusServerEvents.serviceMessage, this.serviceMessageHandler_);
	this.eventBusServer_.on(eventBusServerEvents.messageFromWorker, this.messageHandler_);
	this.eventBusServer_.on(eventBusServerEvents.workerConnected, this.workerConnectHandler_);
	this.eventBusServer_.on(eventBusServerEvents.workerDisconnected, this.workerDisconnectHandler_);
}

/**
 * Parameters related to master config.
 * @type {Object}
 */
MasterNode.parameters = {
	charset: 'charset',
	httpPort: 'master.httpPort',
	eventServicesFolder: 'master.eventServicesFolder',
	sendMessagePerTimeLimit: eventBus.parameters.sendMessagePerTimeLimit,
	sendMessageQueueInterval: eventBus.parameters.sendMessageQueueInterval
};

/**
 * Current config object.
 * @type {Config}
 * @private
 */
MasterNode.prototype.config_ = null;

/**
 * Current master web server.
 * @type {NonCacheableWebServer}
 */
MasterNode.prototype.webServer_ = null;

/**
 * EBS server for cluster communication.
 * @type {EventBusServer}
 * @private
 */
MasterNode.prototype.eventBusServer_ = null;

/**
 * Cluster balancer.
 * @type {Balancer}
 * @private
 */
MasterNode.prototype.balancer_ = null;

/**
 * Current event service provider.
 * @type {EventServiceProvider}
 * @private
 */
MasterNode.prototype.eventServiceProvider_ = null;

/**
 * Current queue for master messages.
 * @type {Array}
 * @private
 */
MasterNode.prototype.eventServiceMessageQueue_ = null;

/**
 * Current event service message send scheduler.
 * @type {Scheduler}
 * @private
 */
MasterNode.prototype.eventServiceMessageSendScheduler_ = null;

/**
 * Message count to send per one time.
 * @type {Number}
 * @private
 */
MasterNode.prototype.sendMessagePerTimeLimit_ = 0;

/**
 * Current worker nodes collection by address.
 * @type {Object}
 * @private
 */
MasterNode.prototype.workersByAddress_ = null;

/**
 * Sends current event service message queue.
 * @private
 */
MasterNode.prototype.sendEventServiceMessageQueueHandler_ = function () {

	if (this.eventServiceMessageQueue_.length === 0) {
		return;
	}

	var workerAddress = this.balancer_.getFreeWorkerId(),
		messagesToSend;

	if (!workerAddress) {
		return;
	}

	messagesToSend = this.eventServiceMessageQueue_.splice(0, this.sendMessagePerTimeLimit_);

	try {
		this.eventBusServer_.sendToWorkerByAddress(messagesToSend, workerAddress);
	} catch (e) {
		this.eventServiceMessageQueue_ = messagesToSend.concat(this.eventServiceMessageQueue_);
	}

};

/**
 * Handles all messages from event services.
 * @param {String} serviceName Name of event service.
 * @param {String} modelName Receiver mode name.
 * @param {Object} data Message object.
 * @private
 */
MasterNode.prototype.eventServiceMessageHandler_ = function (serviceName, modelName, data) {

	this.eventServiceMessageQueue_.push({
		sender: serviceName,
		type: eventBus.messageTypes.data,
		value: {modelName: modelName, data: data}
	});
};

/**
 * Handles all cluster messages.
 * @param {Object} messageObject Incoming message object.
 * @param {String} workerId Message sender id.
 * @param {String} socketAddress Sender socket address string.
 * @private
 */
MasterNode.prototype.messageHandler_ = function (messageObject, workerId, socketAddress) {

	if (!messageObject.type) {
		return;
	}

	var handlerName = util.format('%sHandler_', messageObject.type);

	if (!(handlerName in this)) {
		return;
	}

	this[handlerName](messageObject, workerId, socketAddress);

};

//noinspection JSUnusedGlobalSymbols
/**
 * Handles worker address message.
 * @param {Object} messageObject Incoming message object.
 * @param {String} workerId Message sender ID.
 * @param {String} socketAddress Sender socket address string.
 * @private
 */
MasterNode.prototype.addressHandler_ = function (messageObject, workerId, socketAddress) {
	this.workersByAddress_[socketAddress] = messageObject.value;
};

//noinspection JSUnusedGlobalSymbols,JSUnusedLocalSymbols
/**
 * Handles workers sync and shared messages.
 * @param {Object} messageObject Incoming message object.
 * @param {String} workerId Message sender ID.
 * @private
 */
MasterNode.prototype.sharedHandler_ =
	MasterNode.prototype.syncHandler_ = function (messageObject, workerId) {
		this.eventBusServer_.sendToWorkersExcept(messageObject, workerId);
	};

//noinspection JSUnusedGlobalSymbols
/**
 * Handles worker load message. Worker sends its load estimation.
 * @param {Object} messageObject Incoming message object.
 * @param {String} workerId Message sender ID.
 * @param {String} socketAddress Sender socket address string.
 * @private
 */
MasterNode.prototype.loadHandler_ = function (messageObject, workerId, socketAddress) {
	this.balancer_.setWeight(socketAddress, messageObject.value);
};

//noinspection JSUnusedGlobalSymbols,JSUnusedLocalSymbols
/**
 * Handles data messages from models and route it to event services.
 * @param {Object} messageObject Message from worker.
 * @param {String} workerId Sender worker ID.
 * @private
 */
MasterNode.prototype.dataHandler_ = function (messageObject, workerId) {

	var eventServices = this.eventServiceProvider_.getServices(),
		serviceReceiver;

	if (!messageObject.value || !messageObject.value.serviceName || !eventServices[messageObject.value.serviceName]) {
		return;
	}

	serviceReceiver = eventServices[messageObject.value.serviceName];

	if (!('handleData' in serviceReceiver) || typeof(serviceReceiver.handleData) !== 'function') {
		return;
	}

	serviceReceiver.handleData(messageObject.value.modelName, messageObject.value.data);

};

/**
 * Handles all HTTP requests
 * @param {Object} request HTTP request.
 * @param {Object} response HTTP response.
 * @private
 */
MasterNode.prototype.httpRequestHandler_ = function (request, response) {

	var parsedUrl = url.parse(request.url),
		workerInfo,
		json;

	switch (parsedUrl.pathname) {

		case '/getWorker':
			workerInfo = this.getFreeWorker() || {};
			json = JSON.stringify(workerInfo);

			response.writeHead(200, {
				'Content-Type': 'application/json'
			});
			response.end(json);
			return;

		default:
			response.writeHead(404);
			response.end();
	}
};

/**
 * Handles service messages and log it.
 * @param {Object} messageObject Message object.
 * @private
 */
MasterNode.prototype.serviceMessageHandler_ = function (messageObject) {
	logger.trace(
		util.format('Connection: %s. %s', messageObject.connection, messageObject.message));
};

/**
 * Handler for worker connect event.
 * @param {String} workerAddress Worker address string.
 * @private
 */
MasterNode.prototype.workerConnectHandler_ = function (workerAddress) {
	this.balancer_.addWorker(workerAddress);
	this.workersByAddress_[workerAddress] = {};
};

/**
 * Handler for worker disconnect event.
 * @param {String} workerAddress Worker address string.
 * @private
 */
MasterNode.prototype.workerDisconnectHandler_ = function (workerAddress) {
	this.balancer_.removeWorker(workerAddress);
	delete this.workersByAddress_[workerAddress];
};

/**
 * Starts cluster master.
 */
MasterNode.prototype.start = function () {
	this.eventBusServer_.start();
	this.webServer_.start();
};

/**
 * Gets address of free worker node.
 * @return {String}
 */
MasterNode.prototype.getFreeWorker = function () {
	var freeWorkerId = this.balancer_.getFreeWorkerId();
	return this.workersByAddress_[freeWorkerId];
};