/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var http = require('http'),
	path = require('path'),
	event = require('events'),
	util = require('util'),
	fs = require('fs'),
	Logger = require('../../security/logger.js'),
	logger,
	ModelBase = require('../../mvc/model/model-base.js'),
	ViewBase = require('../../mvc/view/view-base.js'),
	ControllerBase = require('../../mvc/controller/controller-base.js'),
	Factory = require('../../mvc/factory.js'),
	ViewHub = require('../../mvc/view/view-hub.js');

module.exports = LogicSet;

util.inherits(LogicSet, event.EventEmitter);

/**
 * Create new instance of neutrino logic set.
 * @param {Config} config Neutrino config object.
 * @param {WorkerNode} worker Worker which is the owner of this logic set.
 * @constructor
 */
function LogicSet(config, worker) {

	var viewHubEvents = ViewHub.events,
		workerParameters = worker.constructor.parameters,
		workerEvents = worker.constructor.events;

	event.EventEmitter.call(this);

	// bind handlers
	this.modelReloadHandler_ = this.modelReloadHandler_.bind(this);
	this.messageForModelsHandler_ = this.messageForModelsHandler_.bind(this);
	this.clientConnectedHandler_ = this.clientConnectedHandler_.bind(this);
	this.clientDisconnectedHandler_ = this.clientDisconnectedHandler_.bind(this);
	this.httpServerStartedHandler_ = this.httpServerStartedHandler_.bind(this);
	this.sharedMessageHandler_ = this.sharedMessageHandler_.bind(this);
	this.routedNotificationHandler_ = this.routedNotificationHandler_.bind(this);

	this.worker_ = worker;
	this.config_ = config;
	logger = new Logger(config);

	this.models_ = {};
	this.controllers_ = {};
	this.views_ = {};

	this.modelsFolder_ = config.$(workerParameters.mvcModelsFolder);
	this.controllersFolder_ = config.$(workerParameters.mvcControllersFolder);
	this.viewsFolder_ = config.$(workerParameters.mvcViewsFolder);

	this.viewHub_ = new ViewHub(config, worker, worker.sessionManager);
	this.viewHub_.on(viewHubEvents.clientConnected, this.clientConnectedHandler_);
	this.viewHub_.on(viewHubEvents.clientDisconnected, this.clientDisconnectedHandler_);
	this.viewHub_.on(viewHubEvents.httpServerStarted, this.httpServerStartedHandler_);
	this.viewHub_.on(viewHubEvents.routeNotification, this.routedNotificationHandler_);

	this.worker_.on(workerEvents.connectedToMaster, this.modelReloadHandler_);
	this.worker_.on(workerEvents.messageForModels, this.messageForModelsHandler_);
	this.worker_.on(workerEvents.sharedMessage, this.sharedMessageHandler_);

	this.initModels_();
}

/**
 * Enum of logic set events.
 * @enum {String}
 */
LogicSet.events = {
	loaded: 'loaded',
	error: 'error'
};

/**
 * Owner worker.
 * @type {WorkerNode}
 * @private
 */
LogicSet.prototype.worker_ = null;

/**
 * Current config object.
 * @type {Config}
 * @private
 */
LogicSet.prototype.config_ = null;

/**
 * Current server-client view bridge.
 * @type {ViewHub}
 * @private
 */
LogicSet.prototype.viewHub_ = null;

/**
 * Current models folder.
 * @type {String}
 * @private
 */
LogicSet.prototype.modelsFolder_ = '';

/**
 * Current controllers folder.
 * @type {String}
 * @private
 */
LogicSet.prototype.controllersFolder_ = '';

/**
 * Current view folder.
 * @type {String}
 * @private
 */
LogicSet.prototype.viewsFolder_ = '';

/**
 * Hash table of models.
 * @type {Object}
 * @private
 */
LogicSet.prototype.models_ = null;

/**
 * Hash table of controllers.
 * @type {Object}
 * @private
 */
LogicSet.prototype.controllers_ = null;

/**
 * Hash table of views.
 * @type {Object}
 * @private
 */
LogicSet.prototype.views_ = null;

/**
 * Handles client connect event.
 * @private
 */
LogicSet.prototype.clientConnectedHandler_ = function () {
	this.worker_.loadEstimation++;
};

/**
 * Handles client disconnect event.
 * @private
 */
LogicSet.prototype.clientDisconnectedHandler_ = function () {
	this.worker_.loadEstimation--;
};

/**
 * Handles HTTP server start event
 * and set current worker address for clients.
 * @param {String} host HTTP server host.
 * @param {Number} port HTTP server port.
 * @private
 */
LogicSet.prototype.httpServerStartedHandler_ = function (host, port) {
	this.worker_.setCurrentAddress(host, port);
};

/**
 * Handles model shared messages.
 * @param {Object} messageObject Message object.
 * @private
 */
LogicSet.prototype.sharedMessageHandler_ = function (messageObject) {
	if (!messageObject || !messageObject.viewName || !messageObject.sessionId) {
		return;
	}
	this.viewHub_.sendNotification(messageObject.viewName, messageObject.sessionId,
		messageObject.data, true);
};

/**
 * Handles routed messages and send it as a shared message.
 * @param {String} viewName Name of view.
 * @param {String} sessionId User-receiver's session ID.
 * @param {Object} notificationObject Notification object.
 * @private
 */
LogicSet.prototype.routedNotificationHandler_ = function (viewName, sessionId, notificationObject) {
	this.worker_.sendSharedMessage({
		sessionId: sessionId,
		data: notificationObject,
		viewName: viewName
	});
};

/**
 * Emits reload required event in all models.
 * @private
 */
LogicSet.prototype.modelReloadHandler_ = function () {

	for (var key in this.models_) {

		if (!this.models_.hasOwnProperty(key)) {
			return;
		}

		this.models_[key].emit(ModelBase.events.reloadRequired);
	}

};

/**
 * Handle messages for specified model or all models.
 * @param {String} type Type of message.
 * @param {String} sender Name of sender which sends this message
 * @param {Object} value Message object.
 * @private
 */
LogicSet.prototype.messageForModelsHandler_ = function (type, sender, value) {

	var messageHandlerName = util.format('%sMessageHandler', type);

	if (value.modelName && value.modelName in this.models_) {

		if (typeof(this.models_[value.modelName][messageHandlerName]) === 'function') {
			this.models_[value.modelName][messageHandlerName](sender, value.data);
		}
		return;
	}

	for (var modelName in this.models_) {

		if (!this.models_.hasOwnProperty(modelName) ||
			typeof(this.models_[modelName][messageHandlerName]) !== 'function') {
			continue;
		}

		this.models_[modelName][messageHandlerName](sender, value.data);
	}
};

/**
 * Link view and view bridge.
 * @param {String} viewName Name of view to link
 * @private
 */
LogicSet.prototype.linkView_ = function (viewName) {

	var self = this,
		view = this.views_[viewName],
		viewEvents = ViewBase.events,
		viewHubEvents = ViewHub.events;

	view.on(viewEvents.updateValue, function (propertyName, oldValue, newValue, sessionId) {
		self.viewHub_.sendNewValue(viewName, propertyName, oldValue, newValue, sessionId);
	});

	view.on(viewEvents.showResponse, function (response, sessionId, requestId) {
		self.viewHub_.sendResponse(viewName, response, sessionId, requestId);
	});

	view.on(viewEvents.showNotification, function (sessionId, notificationObject) {
		self.viewHub_.sendNotification(viewName, sessionId, notificationObject);
	});

	this.viewHub_.on(viewHubEvents.modelRequest, function (requestViewName, sessionId, requestId) {
		if (requestViewName !== viewName) {
			return;
		}
		view.getModel(sessionId, requestId);
	});

	this.viewHub_.on(viewHubEvents.editRequest,
		function (requestViewName, propertyName, newValue, sessionId, requestId) {
			if (requestViewName !== viewName) {
				return;
			}
			view.setValue(propertyName, newValue, sessionId, requestId);
		});

	this.viewHub_.on(viewHubEvents.invokeRequest, function (requestViewName, methodName, args, sessionId, requestId) {
		if (requestViewName !== viewName) {
			return;
		}
		view.invoke(methodName, args, sessionId, requestId);
	});

	this.viewHub_.on(viewHubEvents.subscribeRequest, function (requestViewName, sessionId, requestId) {
		if (requestViewName !== viewName) {
			return;
		}
		view.subscribe(sessionId, requestId);
	});

	this.viewHub_.on(viewHubEvents.unsubscribeRequest, function (requestViewName, sessionId, requestId) {
		if (requestViewName !== viewName && requestViewName !== '*') {
			return;
		}
		view.unsubscribe(sessionId, requestId);
	});
};

/**
 * Load models and init MVC implementation.
 * @private
 */
LogicSet.prototype.initModels_ = function () {

	var self = this,
		modelEvents = ModelBase.events;

	fs.readdir(this.modelsFolder_, function (error, files) {

		if (error) {
			return;
		}

		files.forEach(function (file) {

			if (path.extname(file) !== '.js') {
				return;
			}

			var modelPath = path.resolve(util.format('%s/%s', self.modelsFolder_, file)),
				controllerPath = path.resolve(util.format('%s/%s', self.controllersFolder_, file)),
				viewPath = path.resolve(util.format('%s/%s', self.viewsFolder_, file)),
				factory = new Factory(self.config_),
				model,
				modelName,
				modelConstructor = require(modelPath),
				controller,
				controllerConstructor,
				view,
				viewConstructor;

			logger.trace(util.format('Model "%s" loaded', modelPath));

			try {
				controllerConstructor = require(controllerPath);
				logger.trace(util.format('Custom controller "%s" loaded', controllerPath));
			} catch (e) {
				controllerConstructor = ControllerBase;
				logger.trace(util.format('Base controller for model "%s" loaded', modelPath));
			}

			try {
				viewConstructor = require(viewPath);
				logger.trace(util.format('Custom view "%s" loaded', viewPath));
			} catch (e) {
				viewConstructor = ViewBase;
				logger.trace(util.format('Base view for model "%s" loaded', modelPath));
			}

			model = factory.createModel(modelConstructor);
			modelName = model.name;
			view = factory.createView(viewConstructor);
			controller = factory.createController(controllerConstructor, model, view);

			model.once(modelEvents.loaded, function () {
				self.emit(LogicSet.events.loaded, {
					name: modelName,
					path: modelPath
				});
			});

			model.on(modelEvents.error, function (error) {
				self.emit(LogicSet.events.error, error);
			});

			model.on(modelEvents.syncRequired, function (message) {
				self.worker_.sendSyncMessage(modelName, message);
			});

			model.on(modelEvents.sentToService, function (serviceName, data) {
				self.worker_.sendDataMessage(modelName, serviceName, data);
			});

			self.models_[modelName] = model;
			self.controllers_[modelName] = controller;
			self.views_[modelName] = view;

			self.linkView_(modelName);

		});
	});
};