/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var util = require('util'),
	uuid = require('node-uuid'),
	events = require('events'),
	eventBus = require('../event-bus/common.js'),
	Logger = require('../../security/logger.js'),
	logger,
	LogicSet = require('./logic-set.js'),
	Scheduler = require('../../common/patterns/scheduler.js'),
	EventBusClient = require('../event-bus/event-bus-client.js'),
	SessionManager = require('../../security/session-manager.js');

module.exports = WorkerNode;

util.inherits(WorkerNode, events.EventEmitter);

/**
 * Create new instance of worker node.
 * @param {Config} config Neutrino config object.
 * @constructor
 */
function WorkerNode(config) {

	events.EventEmitter.call(this);

	this.config_ = config;
	logger = new Logger(config);

	// bind handlers
	this.messageHandler_ = this.messageHandler_.bind(this);
	this.loadEstimationUpdateHandler_ = this.loadEstimationUpdateHandler_.bind(this);
	this.serviceMessageHandler_ = this.serviceMessageHandler_.bind(this);
	this.workerConnectHandler_ = this.workerConnectHandler_.bind(this);
	this.sessionManagerErrorHandler_ = this.sessionManagerErrorHandler_.bind(this);

	this.id = util.format('%d:%s', process.pid, uuid.v1());
	this.host_ = config.$(WorkerNode.parameters.host);
	this.port_ = config.$(WorkerNode.parameters.port);
	this.secure_ = !!config.$(WorkerNode.parameters.sslEnabled);
	this.loadSendInterval_ = config.$(WorkerNode.parameters.loadSendInterval);

	this.sessionManager = new SessionManager(config);
	this.sessionManager.on(SessionManager.events.error, this.sessionManagerErrorHandler_);

	this.logicSet_ = new LogicSet(config, this);

	this.logicSet_.on(LogicSet.events.error,
		function (error) {
			this.emit(WorkerNode.events.error, error);
		}.bind(this));

	this.eventBusClient_ = new EventBusClient(config);

	var eventBusClientEvents = EventBusClient.events;

	this.eventBusClient_.on(eventBusClientEvents.serviceMessage, this.serviceMessageHandler_);
	this.eventBusClient_.on(eventBusClientEvents.messageFromMaster, this.messageHandler_);
	this.eventBusClient_.on(eventBusClientEvents.workerConnected, this.workerConnectHandler_);

	this.loadEstimationUpdateScheduler_ =
		new Scheduler(this.loadSendInterval_, this.loadSendInterval_, this.loadEstimationUpdateHandler_);

	this.loadEstimationUpdateScheduler_.start();
}

/**
 * Enum of worker events.
 * @enum {String}
 */
WorkerNode.events = {
	error: 'error',
	messageForModels: 'messageForModels',
	sharedMessage: 'sharedMessage',
	connectedToMaster: 'connectedToMaster'
};

/**
 * Parameters related to worker node config.
 * @type {Object}
 */
WorkerNode.parameters = {
	host: 'worker.host',
	port: 'worker.port',
	loadSendInterval: 'worker.loadSendInterval',
	sslEnabled: 'worker.ssl.enabled',
	sslKeyPath: 'worker.ssl.keyPath',
	sslCertPath: 'worker.ssl.certPath',
	mvcModelsFolder: 'worker.mvc.modelsFolder',
	mvcControllersFolder: 'worker.mvc.controllersFolder',
	mvcViewsFolder: 'worker.mvc.viewsFolder'
};

/**
 * Current worker ID.
 * @type {String}
 */
WorkerNode.prototype.id = '';

//noinspection JSUnusedGlobalSymbols
/**
 * Load update sending scheduler.
 * @type {Scheduler}
 */
WorkerNode.prototype.loadEstimationUpdateScheduler_ = null;

/**
 * Current logic set of worker.
 * @type {LogicSet}
 * @private
 */
WorkerNode.prototype.logicSet_ = null;

/**
 * EBC instance.
 * @type {EventBusClient}
 * @private
 */
WorkerNode.prototype.eventBusClient_ = null;

/**
 * Interval for load estimation sending.
 * @type {Number}
 */
WorkerNode.prototype.loadSendInterval_ = 0;

/**
 * Current worker node load estimation.
 * @type {Number}
 */
WorkerNode.prototype.loadEstimation = 0;

/**
 * Current session manager instance.
 * @type {SessionManager}
 */
WorkerNode.prototype.sessionManager = null;

/**
 * Current worker node host name.
 * @type {String}
 * @private
 */
WorkerNode.prototype.host_ = '';

/**
 * Current worker node port.
 * @type {Number}
 * @private
 */
WorkerNode.prototype.port_ = '';

/**
 * Is worker client connection secure.
 * @type {Boolean}
 * @private
 */
WorkerNode.prototype.secure_ = false;

/**
 * Send load estimation to master.
 * @private
 */
WorkerNode.prototype.loadEstimationUpdateHandler_ = function () {

	this.eventBusClient_.sendToMaster({
		sender: this.id,
		type: eventBus.messageTypes.load,
		value: this.loadEstimation
	});
};

/**
 * Send load estimation to master.
 * @private
 */
WorkerNode.prototype.addressUpdate_ = function () {

	this.eventBusClient_.sendToMaster({
		sender: this.id,
		type: eventBus.messageTypes.address,
		value: {host: this.host_, port: this.port_, secure: this.secure_}
	});

};

/**
 * Validate incoming message and return true if it's valid.
 * @param {Object} messageObject Message object.
 * @return {Boolean}
 * @private
 */
WorkerNode.prototype.validateMessage_ = function (messageObject) {

	return messageObject &&
		messageObject.type &&
		messageObject.sender &&
		messageObject.sender != this.id;
};

/**
 * Handle all incoming messages.
 * @param {Object} messageObject Incoming message object.
 */
WorkerNode.prototype.messageHandler_ = function (messageObject) {

	if (!this.validateMessage_(messageObject)) {
		return;
	}

	if (messageObject.type === eventBus.messageTypes.shared) {
		this.emit(WorkerNode.events.sharedMessage, messageObject.value);
		return;
	}

	this.emit(WorkerNode.events.messageForModels, messageObject.type, messageObject.sender, messageObject.value);
};

/**
 * Handles session manager error event and throw it up.
 * @param {Error} error Error object.
 * @private
 */
WorkerNode.prototype.sessionManagerErrorHandler_ = function (error) {
	this.emit(WorkerNode.events.error, error);
};

/**
 * Handles service error message and log it.
 * @param {Object} messageObject Message object.
 * @private
 */
WorkerNode.prototype.serviceMessageHandler_ = function (messageObject) {
	logger.trace(util.format('Connection: %s. %s', messageObject.connection, messageObject.message));
};

/**
 * Handles worker connect event and log it.
 * @private
 */
WorkerNode.prototype.workerConnectHandler_ = function () {
	this.addressUpdate_();
	this.loadEstimationUpdateHandler_();
	this.emit(WorkerNode.events.connectedToMaster);
};

/**
 * Send data message to all nodes.
 * @param {Object} message Shared data.
 */
WorkerNode.prototype.sendSharedMessage = function (message) {

	this.eventBusClient_.sendToMaster({
		type: eventBus.messageTypes.shared,
		sender: this.id,
		value: message
	});
};

/**
 * Send synchronization message to master.
 * @param {String} modelName Sender model name.
 * @param {Object} data Synchronization data.
 */
WorkerNode.prototype.sendSyncMessage = function (modelName, data) {

	this.eventBusClient_.sendToMaster({
		type: eventBus.messageTypes.sync,
		sender: this.id,
		value: {modelName: modelName, data: data}
	});
};

/**
 * Send data message to specified event service on master.
 * @param {String} modelName Sender model name.
 * @param {String} serviceName Receiver service name.
 * @param {Object} data Data to send.
 */
WorkerNode.prototype.sendDataMessage = function (modelName, serviceName, data) {

	this.eventBusClient_.sendToMaster({
		type: eventBus.messageTypes.data,
		sender: this.id,
		value: {modelName: modelName, serviceName: serviceName, data: data}
	});
};

/**
 * Set new host and port for incoming client connections.
 * @param {String} host Host address.
 * @param {Number} port Port number.
 */
WorkerNode.prototype.setCurrentAddress = function (host, port) {
	this.host_ = host;
	this.port_ = port;
	this.addressUpdate_();
};

/**
 * Start new worker node.
 */
WorkerNode.prototype.start = function () {
	this.eventBusClient_.connect();
};
