/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var http = require('http'),
	util = require('util'),
	events = require('events');

module.exports = NonCacheableWebServer;

util.inherits(NonCacheableWebServer, events.EventEmitter);

/**
 * Creates new instance of WEB server,
 * which sends "no cache" headers on each response.
 * @param {Number} port Web server port.
 * @constructor
 */
function NonCacheableWebServer(port) {

	events.EventEmitter.call(this);
	this.setMaxListeners(0);

	// bind handlers
	this.httpRequestHandler_ = this.httpRequestHandler_.bind(this);

	this.httpPort_ = port || this.httpPort_;
	this.httpServer_ = http.createServer(this.httpRequestHandler_);
	this.httpServer_.listen(this.httpPort_);
}

/**
 * Event names.
 * @type {Object}
 */
NonCacheableWebServer.prototype.events = {
	onRequest: 'onRequest'
};

//noinspection JSValidateJSDoc
/**
 * HTTP server object.
 * @type {http.Server}
 * @private
 */
NonCacheableWebServer.prototype.httpServer_ = null;

/**
 * Port of HTTP server.
 * @type {Number}
 * @private
 */
NonCacheableWebServer.prototype.httpPort_ = 80;

/**
 * Starts server on port is specified by config.
 */
NonCacheableWebServer.prototype.start = function () {
	this.httpServer_.listen(this.httpPort_);
};

/**
 * Stops server.
 * @param {function} callback Callback after server stops.
 */
NonCacheableWebServer.prototype.stop = function (callback) {

	if (!this.httpServer_) {
		return;
	}

	this.httpServer_.stop(callback);
};

/**
 * Handle all HTTP requests
 * @param {Object} request HTTP request.
 * @param {Object} response HTTP response.
 * @private
 */
NonCacheableWebServer.prototype.httpRequestHandler_ = function (request, response) {

	var now = new Date();

	response.setHeader('Server', 'neutrino');
	response.setHeader('Access-Control-Allow-Origin', '*');
	response.setHeader('Expires', 'Thu, 01 Jan 1970 00:00:01 GMT');
	response.setHeader('Last-Modified', now.toGMTString());
	response.setHeader('Date', now.toGMTString());
	response.setHeader('Cache-Control', 'no-cache, must-revalidate');
	response.setHeader('Pragma', 'no-cache');

	this.emit(this.events.onRequest, request, response);
};

