/*
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var Scheduler = require('./scheduler.js');

module.exports = {

	/**
	 * Time-divided 'for' loop.
	 * @param {Object} state Any init state of loop.
	 * @param {function} predicate Predicate which controls loop interruption.
	 * @param {function} iteration Function which change state for next iteration.
	 * @param {function} action Loop body action.
	 */
	for: function (state, predicate, iteration, action) {

		var curState = state;
		var scheduler = new Scheduler(0, 0, function () {
			if (!predicate(curState)) {
				scheduler.stop();
				return;
			}

			action(curState);
			curState = iteration(curState);
		});
		scheduler.start();
	},

	/**
	 * Time-divided 'for' loop with counter.
	 * @param {Number} initValue Init counter value.
	 * @param {Number} limit Counter limit (not including).
	 * @param {function} action Loop body action.
	 */
	countedFor: function (initValue, limit, action) {
		this.for(initValue,
			function (counter) {
				return counter < limit;
			},
			function (counter) {
				return counter + 1;
			}, action);
	},

	/**
	 * Time-divided 'while' loop.
	 * @param {function} predicate Predicate which controls loop interruption.
	 * @param {function} action Loop body action.
	 */
	while: function (predicate, action) {
		var scheduler = new Scheduler(0, 0, function () {
			if (!predicate()) {
				scheduler.stop();
				return;
			}

			action();
		});
		scheduler.start();
	}

};
