/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

module.exports = Scheduler;

/**
 * Creates new instance of scheduler.
 * @param {Number} delay Start delay.
 * @param {Number} interval Iteration interval.
 * @param {function} callback Scheduler callback.
 * @constructor
 */
function Scheduler(delay, interval, callback) {

	this.delay_ = delay || this.delay_;
	this.interval_ = interval || this.interval_;
	this.callback_ = typeof(callback) === 'function' ? callback : this.callback_();
	this.runCallback_ = this.runCallback_.bind(this);
}

/**
 * Scheduler interval.
 * @type {Number}
 * @private
 */
Scheduler.prototype.interval_ = 0;

/**
 * Scheduler first run delay.
 * @type {Number}
 * @private
 */
Scheduler.prototype.delay_ = 0;

/**
 * Default scheduler callback.
 * @private
 */
Scheduler.prototype.callback_ = function () {

};

/**
 * Current timeout handle.
 * @type {Number}
 * @private
 */
Scheduler.prototype.timeoutHandle_ = null;

/**
 * Current scheduler state.
 * @type {Boolean}
 * @private
 */
Scheduler.prototype.isStarted_ = false;

/**
 * Runs next iteration of scheduler call.
 * @private
 */
Scheduler.prototype.runCallback_ = function () {

	if (!this.isStarted_) {
		return;
	}

	if (this.timeoutHandle_) {
		clearTimeout(this.timeoutHandle_);
	}

	this.callback_();

	if (this.interval_ >= 0) {

		this.timeoutHandle_ = setTimeout(
			this.runCallback_,
			this.interval_);

	} else {
		this.stop();
	}
};

/**
 * Starts scheduler.
 */
Scheduler.prototype.start = function () {

	if (this.isStarted_) {
		return;
	}

	this.isStarted_ = true;

	if (this.delay_ > 0) {

		this.timeoutHandle_ = setTimeout(
			this.runCallback_,
			this.delay_);

	} else {
		this.runCallback_();
	}
};

/**
 * Stops scheduler.
 */
Scheduler.prototype.stop = function () {

	if (this.timeoutHandle_) {
		clearTimeout(this.timeoutHandle_);
	}

	this.isStarted_ = false;
};