/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */
/**
 * @fileoverview neutrino config storage implementation.
 * This depend on
 * path
 * @author denis.rechkunov@gmail.com (Denis Rechkunov)
 */

var path = require('path'),
	defaults = require('../defaults.json'),
	fs = require('fs');

module.exports = Config;

/**
 * Read-only config storage.
 * @constructor
 * @param {string | object} config Path to config file or config object.
 */
function Config(config) {

	this.configObj_ = {};

	var existsPredicate = fs.existsSync || path.existsSync,
		configPath;

	// if config is a string then load file from path
	// if it's already object then save it
	// else it's empty object
	if (config && typeof(config) === 'string') {

		if (!existsPredicate(config)) {
			throw new Error('Config "' + config + '" not found!');
		}

		configPath = path.resolve(config);
		this.configObj_ = require(configPath); // user-defined config

	} else if (config && typeof(config) === 'object') {
		this.configObj_ = config;
	}
}

/**
 * Protected config object.
 * @type {object}
 * @private
 */
Config.prototype.configObj_ = null;

/**
 * Function for fast accessing to config parameter.
 * @param {string} key Parameter name.
 */
Config.prototype.$ = function (key) {
	return this.getValue(key);
};

/**
 * Searches for config parameter value by key.
 * @param key {String} Parameter key presented by string divided by dots.
 * @return {*}
 */
Config.prototype.getValue = function (key) {

	var parts = key.split('.'),
		currentPart,
		currentSection = this.configObj_,
		currentDefaultSection = defaults;

	do {
		currentPart = parts.shift();

		if (currentSection != null) {
			currentSection = (currentPart in currentSection) ?
				currentSection[currentPart] : null;
		}

		if (currentDefaultSection != null) {
			currentDefaultSection = (currentPart in currentDefaultSection) ?
				currentDefaultSection[currentPart] : null;
		}

	} while (parts.length > 0 && !(currentDefaultSection == null && currentSection == null));

	return currentSection == null ? currentDefaultSection : currentSection;

};