/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */
var mongodb = require('mongodb'),
	util = require('util'),
	events = require('events'),
	singleton = require('../common/patterns/singleton.js');

module.exports = singleton(DatabaseProvider);

util.inherits(DatabaseProvider, events.EventEmitter);

/**
 * Create new instance of database provider.
 * @param {Config} config Neutrino config object.
 * @constructor
 */
function DatabaseProvider(config) {

	events.EventEmitter.call(this);
	this.setMaxListeners(0);

	var host = config.$(DatabaseProvider.parameters.host),
		port = config.$(DatabaseProvider.parameters.port),
		name = config.$(DatabaseProvider.parameters.name),
		options = config.$(DatabaseProvider.parameters.options),
		serverOptions = config.$(DatabaseProvider.parameters.serverOptions),
		user = config.$(DatabaseProvider.parameters.user),
		password = config.$(DatabaseProvider.parameters.password),
		server = new mongodb.Server(host, port, serverOptions);

	this.openHandler_ = this.openHandler_.bind(this, user, password);
	this.authHandler_ = this.authHandler_.bind(this);
	this.closeHandler_ = this.closeHandler_.bind(this);

	this.db_ = new mongodb.Db(name, server, options);
	this.db_.on('close', this.closeHandler_);
	this.db_.on('error', this.errorHandler_);
	this.db_.open(this.openHandler_);

}

/**
 * Enum of database provider events.
 * @enum {String}
 */
DatabaseProvider.events = {
	open: 'open',
	close: 'close',
	auth: 'auth',
	error: 'error'
};

/**
 * Database related parameters.
 * @type {Object}
 */
DatabaseProvider.parameters = {
	host: 'database.host',
	port: 'database.port',
	user: 'database.user',
	password: 'database.password',
	name: 'database.name',
	options: 'database.options',
	serverOptions: 'database.serverOptions'
};

//noinspection JSValidateJSDoc
/**
 * Current instance of database, which provider uses.
 * @type {mongodb.Db}
 * @private
 */
DatabaseProvider.prototype.db_ = null;

/**
 * Is provider pass a database authentication.
 * @type {Boolean}
 * @private
 */
DatabaseProvider.prototype.isAuth_ = false;

//noinspection JSValidateJSDoc
/**
 * Current client for opened database.
 * @type {mongodb.DbClient}
 * @private
 */
DatabaseProvider.prototype.dbClient_ = null;

/**
 * Handle authentication event.
 * @param {Error} error Error object if the error was.
 * @param {Object} result Authentication result object.
 * @private
 */
DatabaseProvider.prototype.authHandler_ = function (error, result) {
	if (error) {
		this.errorHandler_(error);
	}
	this.isAuth_ = result;
	this.emit(DatabaseProvider.events.auth);
};

//noinspection JSValidateJSDoc
/**
 * Handle database open event and start authentication.
 * @param {String} user Database username.
 * @param {String} password User password.
 * @param {Error} error Error object if error was during database opening.
 * @param {mongodb.DbClient} client Client for opened database.
 * @private
 */
DatabaseProvider.prototype.openHandler_ = function (user, password, error, client) {

	if (error) {
		this.errorHandler_(error);
	}

	this.dbClient_ = client;
	this.db_.authenticate(user, password, this.authHandler_);
	this.emit(DatabaseProvider.events.open);
};

/**
 * Handle database close event.
 * @private
 */
DatabaseProvider.prototype.closeHandler_ = function () {
	this.dbClient_ = null;
	this.isAuth_ = false;
	this.emit(DatabaseProvider.events.close);
};

/**
 * Handle errors and provide it to next level.
 * @param {Error} error Error object.
 * @private
 */
DatabaseProvider.prototype.errorHandler_ = function (error) {
	this.emit(DatabaseProvider.events.error, error);
};

//noinspection JSValidateJSDoc
/**
 * Get specified collection for query execution.
 * @param {String} name Collection name.
 * @param {function(mongodb.Collection)} callback Callback to handle requested collection.
 */
DatabaseProvider.prototype.getCollection = function (name, callback) {

	var resultHandler = function () {
		var collection = new mongodb.Collection(this.dbClient_, name);
		callback(collection);
	}.bind(this);

	if (this.dbClient_ && this.isAuth_) {
		resultHandler();
	}

	this.once(DatabaseProvider.events.error, function () {
		callback(null);
	});

	this.once(DatabaseProvider.events.auth, function () {
		resultHandler();
	});
};

/**
 * Close current database.
 */
DatabaseProvider.prototype.close = function () {
	this.db_.close();
};