/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var util = require('util'),
	events = require('events'),
	mvc = require('../common.js'),
	ModelBase = require('../model/model-base.js'),
	ViewBase = require('../view/view-base.js'),
	Barrier = require('../../common/patterns/barrier.js');

module.exports = ControllerBase;

util.inherits(ControllerBase, events.EventEmitter);

/**
 * Implements all base controller actions and validate methods.
 * @param {Config} config Object with current configuration.
 * @param {ModelBase} model Controller's model.
 * @param {ViewBase} view View connected with controller.
 * @constructor
 */
function ControllerBase(config, model, view) {

	var self = this;

	events.EventEmitter.call(self);
	self.setMaxListeners(0);

	self.model_ = model;
	self.view_ = view;
	self.config_ = config;

	self.subscribers_ = {};

	self.on(ControllerBase.events.error, function (error, sessionId, requestId) {
		if (!sessionId) {
			return;
		}
		self.view_.showResponse({
			success: false,
			error: error
		}, sessionId, requestId);
	});

	var viewEvents = ViewBase.events,
		modelEvents = ModelBase.events;

	self.view_.on(viewEvents.setValue, function (propertyName, newValue, sessionId, requestId) {
		self.setValue(propertyName, newValue, sessionId, requestId);
	});

	self.view_.on(viewEvents.subscribe, function (sessionId, requestId) {
		self.subscribe(sessionId, requestId);
	});

	self.view_.on(viewEvents.unsubscribe, function (sessionId, requestId) {
		self.unsubscribe(sessionId, requestId);
	});

	self.view_.on(viewEvents.modelRequest, function (sessionId, requestId) {
		self.getModel(sessionId, requestId);
	});

	self.view_.on(viewEvents.invoke, function (methodName, args, sessionId, requestId) {
		self.invoke(methodName, args, sessionId, requestId);
	});

	self.model_.on(modelEvents.changed, function (propertyName, oldValue, newValue) {
		self.modelUpdateHandler(propertyName, oldValue, newValue);
	});

	self.model_.on(modelEvents.notify, function (sessionId, notificationObject) {
		self.view_.showNotification(sessionId, notificationObject);
	});

	self.model_.on(modelEvents.error, function (error) {
		self.emit(ControllerBase.events.error, error);
	});

}

/**
 * Enum of controller events.
 * @enum {String}
 */
ControllerBase.events = {
	error: 'error'
};

/**
 * Model of controller.
 * @type {ModelBase}
 */
ControllerBase.prototype.model_ = null;

/**
 * View connected with controller.
 * @type {ViewBase}
 */
ControllerBase.prototype.view_ = null;

//noinspection JSUnusedGlobalSymbols
/**
 * Current config object.
 * @type {Config}
 * @private
 */
ControllerBase.prototype.config_ = null;

/**
 * Current subscribers hash table.
 * @type {Object}
 * @private
 */
ControllerBase.prototype.subscribers_ = null;

/**
 * Set new value for specified property of model.
 * @param {String} propertyName Property name to update.
 * @param {*} newValue New value of property.
 * @param {String} sessionId User session ID.
 * @param {String} requestId User request ID.
 */
ControllerBase.prototype.setValue = function (propertyName, newValue, sessionId, requestId) {

	var self = this,
		execute = function () {

			if (!(propertyName in self.model_)) {
				self.emit(ControllerBase.events.error, Error('No property with such name'), sessionId, requestId);
				return;
			}

			if (mvc.privateRegExp.test(propertyName)) {
				self.emit(ControllerBase.events.error, Error('Can not get or set private properties'), sessionId,
					requestId);
				return;
			}

			var setValidatorName = util.format(mvc.propertySetValidatorFormat, propertyName);

			// use set property validator if it exists
			if (setValidatorName in self) {

				self[setValidatorName](newValue, sessionId, function (error) {

					if (error) {
						self.emit(ControllerBase.events.error, error, sessionId, requestId);
					} else {
						self.model_[propertyName].$(newValue);
						self.view_.showResponse({success: true}, sessionId, requestId);
					}

				});

			} else {
				self.model_[propertyName].$(newValue);
				self.view_.showResponse({success: true}, sessionId, requestId);
			}
		};

	// use model access validator if it exists.
	if (mvc.modelAccessValidatorName in self) {
		self[mvc.modelAccessValidatorName](sessionId, function (error) {
			if (error) {
				self.emit(ControllerBase.events.error, error, sessionId, requestId);
			} else {
				execute();
			}
		});
	} else {
		execute();
	}
};

//noinspection JSUnusedGlobalSymbols
/**
 * Invoke specified model method.
 * @param {String} methodName Name of model method.
 * @param {Array} args Method arguments.
 * @param {String} sessionId User session ID.
 * @param {String} requestId User request ID.
 */
ControllerBase.prototype.invoke = function (methodName, args, sessionId, requestId) {

	var self = this,
		execute = function () {

			if (!(methodName in self.model_) || typeof(self.model_[methodName]) !== 'function') {
				self.emit(ControllerBase.events.error, new Error('No method with such name'), sessionId, requestId);
				return;
			}

			if (mvc.privateRegExp.test(methodName)) {
				self.emit(ControllerBase.events.error, new Error('Can not invoke private methods'), sessionId,
					requestId);
				return;
			}

			var invokeValidatorName = util.format(mvc.methodInvokeValidatorFormat, methodName),

				methodCallback = function (result) {
					self.view_.showResponse({success: true, result: result}, sessionId, requestId);
				},
				methodArgs = args.concat([methodCallback]),
				validatorCallback = function (error) {

					if (error) {
						self.emit(ControllerBase.events.error, error, sessionId, requestId);
					} else {
						self.model_[methodName].apply(self.model_, methodArgs);
					}
				},
				validatorArgs = [sessionId, validatorCallback].concat(args);

			if (invokeValidatorName in self) {
				self[invokeValidatorName].apply(self, validatorArgs);
			} else {
				self.model_[methodName].apply(self.model_, methodArgs);
			}

		};

	// use model access validator if it exists.
	if (mvc.modelAccessValidatorName in self) {
		self[mvc.modelAccessValidatorName](sessionId, function (error) {
			if (error) {
				self.emit(ControllerBase.events.error, error, sessionId, requestId);
			} else {
				execute();
			}
		});
	} else {
		execute();
	}
};

/**
 * Get model object with public properties.
 * @param {String} sessionId User session ID.
 * @param {String} requestId User request ID.
 */
ControllerBase.prototype.getModel = function (sessionId, requestId) {

	var self = this,
		modelToSend = {name: self.model_.name, methods: [], properties: {}},
		execute = function () {

			var modelObject = self.model_.serialize(),
				barrier = new Barrier();

			// check all methods for accessibility
			modelObject.methods.forEach(function (methodName) {
				if (mvc.privateRegExp.test(methodName)) {
					return;
				}
				modelToSend.methods.push(methodName);
			});

			// check all properties for accessibility
			for (var key in modelObject.properties) {

				if (!modelObject.properties.hasOwnProperty(key)) {
					continue;
				}

				if (mvc.privateRegExp.test(key)) {
					continue;
				}

				var getValidatorName = util.format(mvc.propertyGetValidatorFormat, key);

				if (getValidatorName in self) {
					// because key and getValidatorName are mutable we must use immediate function
					(function (currentKey, validatorName) {

						barrier.register(function (notifier) {

							// call validator as a barrier task
							self[validatorName](sessionId, function (error) {
								if (!error) {
									modelToSend.properties[currentKey] = modelObject.properties[currentKey];
								}
								notifier(error);
							});

						});

					})(key, getValidatorName);

				} else {
					modelToSend.properties[key] = modelObject.properties[key];
				}
			}

			barrier.execute(function (results) {
				self.view_.showResponse({success: true, model: modelToSend}, sessionId, requestId);
			});
		};

	if (mvc.modelAccessValidatorName in self) {

		self[mvc.modelAccessValidatorName](sessionId, function (error) {
			if (error) {
				self.emit(ControllerBase.events.error, error, sessionId, requestId);
			} else {
				execute();
			}
		});

	} else {
		execute();
	}

};

/**
 * Send to all subscribers update value message.
 * @param {String} propertyName Name of model property.
 * @param {*} oldValue Old property value.
 * @param {*} newValue New property value.
 */
ControllerBase.prototype.modelUpdateHandler = function (propertyName, oldValue, newValue) {

	var self = this;
	if (mvc.privateRegExp.test(propertyName)) {
		return;
	}

	var getValidatorName = util.format(mvc.propertyGetValidatorFormat, propertyName);

	for (var sessionId in self.subscribers_) {

		if (!self.subscribers_.hasOwnProperty(sessionId)) {
			continue;
		}

		if (getValidatorName in self) {
			// because sessionId is mutable we must use immediate function
			(function (sid) {
				self[getValidatorName](sessionId, function (error) {
					if (error) {
						self.emit(ControllerBase.events.error, error, sid)
					} else {
						self.view_.updateValue(propertyName, oldValue, newValue, sid);
					}
				});
			})(sessionId);
		} else {
			self.view_.updateValue(propertyName, oldValue, newValue, sessionId);
		}
	}

};

//noinspection JSUnusedLocalSymbols
/**
 * Subscribe user to model update.
 * @param {String} sessionId User session ID.
 * @param {String} requestId User session ID>
 */
ControllerBase.prototype.subscribe = function (sessionId, requestId) {

	var self = this;

	if (!sessionId) {
		return;
	}

	self.subscribers_[sessionId] = true;

	self.view_.showResponse({
		success: self.subscribers_[sessionId]
	}, sessionId, requestId);

};

//noinspection JSUnusedLocalSymbols
/**
 * Unsubscribe user from model update.
 * @param {String} sessionId User session ID.
 * @param {String} requestId User request ID.
 */
ControllerBase.prototype.unsubscribe = function (sessionId, requestId) {

	var self = this;

	if (!sessionId) {
		return;
	}

	var result = delete self.subscribers_[sessionId];

	self.view_.showResponse({
		success: result
	}, sessionId, requestId);
};
