/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var util = require('util'),
	ModelBase = require('./model/model-base.js'),
	ViewBase = require('./view/view-base.js'),
	ControllerBase = require('./controller/controller-base.js');

module.exports = Factory;

function Factory(config) {
	this.config_ = config;
}

/**
 * Current neutrino config.
 * @type {Config}
 * @private
 */
Factory.prototype.config_ = null;

/**
 * Create new instance of model is inherited from ModelBase.
 * @param {function} customModelConstructor Custom model constructor.
 * @return {*}
 */
Factory.prototype.createModel = function (customModelConstructor) {

	if (customModelConstructor === ModelBase) {
		throw new Error('Model base instance could not be created');
	}

	var constructor = this.createInheritedConstructor(customModelConstructor, ModelBase);

	return new constructor(this.config_);
};

/**
 * Create new instance of controller is inherited from ControllerBase.
 * @param {function} customControllerConstructor Custom controller constructor.
 * @param {*} model Model object is inherited from ModelBase.
 * @param {*} view View object is inherited from ViewBase.
 * @return {*}
 */
Factory.prototype.createController = function (customControllerConstructor, model, view) {

	if (customControllerConstructor === ControllerBase) {
		return new ControllerBase(this.config_, model, view);
	}

	var constructor = this.createInheritedConstructor(customControllerConstructor, ControllerBase);

	return new constructor(this.config_, model, view);
};

/**
 * Create new instance of view is inherited from ViewBase.
 * @param {function} customViewConstructor Custom controller constructor.
 */
Factory.prototype.createView = function (customViewConstructor) {

	if (customViewConstructor === ViewBase) {
		return new ViewBase(this.config_);
	}

	var constructor = this.createInheritedConstructor(customViewConstructor, ViewBase);

	return new constructor(this.config_);
};

/**
 * Creates inherited constructor from existed child constructor.
 * @param {function} childConstructor Child constructor.
 * @param {function} superConstructor Super constructor.
 * @return {function}
 */
Factory.prototype.createInheritedConstructor = function (childConstructor, superConstructor) {

	var childPrototype = childConstructor.prototype;

	util.inherits(childConstructor, superConstructor);
	this.mergePrototypes_(childConstructor.prototype, childPrototype);

	return childConstructor;
};

/**
 * Copy all properties from child constructor
 * prototype to inheritance result prototype.
 * @param {Object} prototype Inheritance result prototype.
 * @param {Object} childPrototype Child constructor prototype.
 * @private
 */
Factory.prototype.mergePrototypes_ = function (prototype, childPrototype) {
	for (var key in childPrototype) {
		if (!childPrototype.hasOwnProperty(key)) {
			continue;
		}
		prototype[key] = childPrototype[key];
	}
};

