/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var util = require('util'),
	events = require('events'),
	Property = require('./property.js'),
	DatabaseProvider = require('../../io/database-provider.js');

module.exports = ModelBase;

util.inherits(ModelBase, events.EventEmitter);

/**
 * Create new instance of neutrino base model.
 * @param {Config} config Neutrino config object.
 * @constructor
 */
function ModelBase(config) {

	var self = this;

	ModelBase.super_.call(self);
	self.setMaxListeners(0);

	self.modelsCollectionName_ = config.$(ModelBase.parameters.mvcModelsCollectionName);
	self.config_ = config;
	self.name = self.constructor.name;
	self.propertyNameList_ = [];
	self.methodNameList_ = [];

	self.on(ModelBase.events.changed, function (propertyName, oldValue, newValue, syncRequired) {
		self.changeHandler_(propertyName, oldValue, newValue, syncRequired);
	});

	self.wrapProperties_();

	self.dbProvider_ = new DatabaseProvider(config);

	self.on(ModelBase.events.storageReady, function () {
		self.isStorageReady_ = true;
	});

	self.on(ModelBase.events.reloadRequired, function () {
		if (!self.isModelLoaded_) {
			return;
		}

		self.loadFromStorage_();
	});

	self.on(ModelBase.events.loaded, function () {
		self.isModelLoaded_ = true;
	});

	self.dbProvider_.getCollection(self.modelsCollectionName_, function (collection) {
		self.storage_ = collection;
		self.emit(ModelBase.events.storageReady);
	});

	self.loadFromStorage_();
}

/**
 * Enum of model events.
 * @enum {String}
 */
ModelBase.events = {
	changed: 'changed',
	loaded: 'loaded',
	saved: 'saved',
	propertySaved: 'propertySaved',
	dataFromService: 'dataFromService',
	error: 'error',
	storageReady: 'storageReady',
	syncRequired: 'syncRequired',
	reloadRequired: 'reloadRequired',
	sentToService: 'sentToService',
	notify: 'notify'
};

/**
 * Parameters related to model.
 * @type {Object}
 */
ModelBase.parameters = {
	mvcModelsCollectionName: 'worker.mvc.modelsCollectionName'
};

/**
 * Current models collection name.
 * @type {String}
 * @private
 */
ModelBase.prototype.modelsCollectionName_ = '';

/**
 * Current database provider.
 * @type {DatabaseProvider}
 * @private
 */
ModelBase.prototype.dbProvider_ = null;

//noinspection JSValidateJSDoc
/**
 * Current models storage.
 * @type {mongodb.Collection}
 * @private
 */
ModelBase.prototype.storage_ = null;

/**
 * Id of model object in storage.
 * @type {String}
 * @private
 */
ModelBase.prototype.idInStorage_ = '';

/**
 * Is models storage ready.
 * @type {Boolean}
 * @private
 */
ModelBase.prototype.isStorageReady_ = false;

/**
 * Is model loaded from storage.
 * @type {Boolean}
 * @private
 */
ModelBase.prototype.isModelLoaded_ = false;

//noinspection JSUnusedGlobalSymbols
/**
 * Current neutrino config object.
 * @type {Config}
 * @private
 */
ModelBase.prototype.config_ = null;

/**
 * Current model name.
 * @type {String}
 */
ModelBase.prototype.name = '';

/**
 * Current model state revision.
 * @type {Number}
 */
ModelBase.prototype.revision_ = 0;

/**
 * Current list of property keys.
 * @type {Array}
 * @private
 */
ModelBase.prototype.propertyNameList_ = null;

/**
 * Current list of method keys.
 * @type {Array}
 * @private
 */
ModelBase.prototype.methodNameList_ = null;

/**
 * Serialize model to object.
 * @return {Object}
 */
//noinspection JSUnusedGlobalSymbols
ModelBase.prototype.serialize = function () {

	var self = this,
		propertyNames = self.getPropertyNames(),
		modelObject = {
			name: self.name,
			revision: self.revision_,
			properties: {},
			methods: self.getMethodNames()
		};

	if (self.idInStorage_) {
		modelObject._id = self.idInStorage_;
	}

	propertyNames.forEach(function (propertyName) {
		modelObject.properties[propertyName] = self[propertyName].$();
	});

	return modelObject;
};

/**
 * Deserialize model from object.
 * @param {Object} modelObject Model object which describes properties.
 */
ModelBase.prototype.deserialize = function (modelObject) {

	if (!modelObject || !('properties' in modelObject) || !('methods' in modelObject)) {
		return;
	}

	this.idInStorage_ = modelObject._id;
	this.revision_ = modelObject.revision;

	var self = this,
		syncPackage = self.createSyncPackage_(),
		properties = modelObject.properties,
		oldValue;

	for (var key in properties) {

		if (!properties.hasOwnProperty(key)) {
			continue;
		}

		if (!(key in self) || !(self[key] instanceof Property)) {
			continue;
		}

		oldValue = self[key].$();

		syncPackage.properties.push({name: key, value: properties[key]});

		if (oldValue === properties[key]) {
			continue;
		}

		self[key].set(properties[key], false);
	}

	self.emit(ModelBase.events.syncRequired, syncPackage);
};

//noinspection JSUnusedGlobalSymbols
/**
 * Handle data messages.
 * @param {String} sender Sender ID.
 * @param {Object} data Data of message.
 */
ModelBase.prototype.dataMessageHandler = function (sender, data) {

	var self = this;
	self.emit(ModelBase.events.dataFromService, sender, data);

};

/**
 * Handle synchronization messages.
 * @param {String} sender Sender ID.
 * @param {Array|Object} data Synchronization data.
 */
ModelBase.prototype.syncMessageHandler = function (sender, data) {

	var self = this,
		oldValue;

	if (data.revision <= self.revision_) {
		return;
	}

	data.properties.forEach(function (changeObject) {

		var key = changeObject.name;
		if (!(key in self) || !(self[key] instanceof Property)) {
			return;
		}

		oldValue = self[key].$();

		if (oldValue === changeObject.value) {
			return;
		}

		self[key].set(changeObject.value, false);
		self.emit(ModelBase.events.changed, key, oldValue, changeObject.value, false);
	});

	self.revision_ = data.revision;
};

/**
 * Handle all model properties changes and send sync messages.
 * @param {String} propertyName Model property name.
 * @param {*} oldValue Old value of property.
 * @param {*} newValue New value of property.
 * @param {Boolean} syncRequired Is sync required (optional).
 * @private
 */
ModelBase.prototype.changeHandler_ = function (propertyName, oldValue, newValue, syncRequired) {

	var self = this;

	if (syncRequired === undefined || syncRequired === true) {

		self.once(ModelBase.events.propertySaved, function () {

			var syncPackage = self.createSyncPackage_();
			self.propertyNameList_.forEach(function (name) {
				syncPackage.properties.push({name: name, value: self[name].$()});
			});

			self.emit(ModelBase.events.syncRequired, syncPackage);
		});

		self.editInStorage_(propertyName, newValue);
	}
};

/**
 * Create synchronization message.
 * @return {Object}
 * @private
 */
ModelBase.prototype.createSyncPackage_ = function () {

	return{
		revision: this.revision_,
		properties: []
	};
};

/**
 * Apply model property and method configuration with default values.
 * @private
 */
ModelBase.prototype.wrapProperties_ = function () {

	var self = this,
		isSerializable,
		prototype = self.constructor.prototype;

	for (var key in prototype) {

		if (!prototype.hasOwnProperty(key)) {
			continue;
		}

		isSerializable = false;

		switch (typeof prototype[key]) {

			case 'number':
			case 'boolean':
			case 'string':
				isSerializable = true;
				break;
			case 'object':
				isSerializable = (typeof prototype[key] instanceof Date);
				break;

			case 'function':
				self.methodNameList_.push(key);
				break;
		}

		if (isSerializable) {

			self.propertyNameList_.push(key);

			self[key] = new Property(key, prototype[key]);
			self[key].on(Property.events.changed, function (name, oldValue, newValue) {
				self.emit(ModelBase.events.changed, name, oldValue, newValue);
			});
		}
	}
};

/**
 * Load model object from models storage.
 * @private
 */
ModelBase.prototype.loadFromStorage_ = function () {

	var self = this,
		execute = function () {
			self.storage_.findOne(
				{name: self.name},
				function (error, modelObject) {
					if (error) {
						self.emit(ModelBase.events.error, error);
						return;
					}

					if (!modelObject) {
						self.saveToStorage_();
						self.once(ModelBase.events.saved, function () {
							self.emit(ModelBase.events.loaded);
						});
					} else {
						self.deserialize(modelObject);
						self.emit(ModelBase.events.loaded);
					}
				});
		};

	if (self.isStorageReady_) {
		execute();
	} else {
		self.once(ModelBase.events.storageReady, execute);
	}

};

/**
 * Save current model state to models storage.
 * @private
 */
ModelBase.prototype.saveToStorage_ = function () {

	var self = this,
		execute = function () {
			self.storage_.save(
				self.serialize(),
				{safe: true},
				function (error, insertedObjectOrCount) {

					if (error) {
						self.emit(ModelBase.events.error, error);
						return;
					}
					if (typeof insertedObjectOrCount === 'object') {
						self.idInStorage_ = insertedObjectOrCount._id;
						self.emit(ModelBase.events.saved);
					}
				});
		};

	if (self.isStorageReady_) {
		execute();
	} else {
		self.once(ModelBase.events.storageReady, execute);
	}

};

/**
 * Edit model value in saved model storage.
 * @param {String} propertyName Name of property.
 * @param {*} newValue New value of property.
 * @private
 */
ModelBase.prototype.editInStorage_ = function (propertyName, newValue) {

	var setParameters = {
		properties: {}
	};

	setParameters.properties[propertyName] = newValue;

	var self = this,
		execute = function () {
			// http://docs.mongodb.org/manual/reference/command/findAndModify/
			// it says "This command obtains a write lock on the affected database and will block other operations until it has completed"
			self.storage_.findAndModify(
				{_id: self.idInStorage_},
				{sort: [
					['revision', 'desc']
				]},
				{
					$set: setParameters,
					$inc: {revision: 1}
				},
				{
					safe: true,
					new: true
				},
				function (error, updatedObject) {
					if (error) {
						self.emit(ModelBase.events.error, error);
						return;
					}

					if (!updatedObject) {
						return;
					}

					if (updatedObject.revision - self.revision_ > 1) {
						self.emit(ModelBase.events.reloadRequired);
					} else {
						self.revision_ = updatedObject.revision;
					}

					self.emit(ModelBase.events.propertySaved, propertyName, newValue);
				});
		};

	if (self.isModelLoaded_) {
		execute();
	} else {
		self.once(ModelBase.events.loaded, execute);
	}

};

/**
 * Gets model's method names.
 * @return {Array}
 */
ModelBase.prototype.getMethodNames = function () {
	return this.methodNameList_;
};

/**
 * Gets model's property names.
 * @return {Array}
 */
ModelBase.prototype.getPropertyNames = function () {
	return this.propertyNameList_;
};