/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var util = require('util'),
	events = require('events'),
	ModelBase = require('./model-base.js');

module.exports = TrackableModel;

util.inherits(TrackableModel, ModelBase);

/**
 * Create new instance of trackable model.
 * @param {Config} config Neutrino config object.
 * @param {String} name Name of model.
 * @param {Object} propertyConfig Object which describes model properties.
 * @param {String} trackCollectionName Name of collection
 * which will be a container for track objects.
 * @constructor
 */
function TrackableModel(config, name, propertyConfig, trackCollectionName) {

	var self = this;

	if (!trackCollectionName || typeof(trackCollectionName) != 'string') {
		throw new Error('Track collection name must be specified');
	}

	TrackableModel.super_.call(self, config, name, propertyConfig);
	self.trackCollectioName_ = trackCollectionName;

	self.dbProvider_.getCollection(self.trackCollectionName_, function (collection) {
		self.trackStorage_ = collection;
		self.emit(TrackableModel.events.trackStorageReady);
	});

	self.on(TrackableModel.events.trackStorageReady, function () {
		self.isTrackStorageReady_ = true;
	});

	self.on(ModelBase.events.changed, function () {
		self.changeTrackHandler_();
	});
}

/**
 * Enum of trackable model events.
 * @enum {String}
 */
TrackableModel.events = {
	trackStorageReady: 'trackStorageReady'
};

/**
 * Is model track storage ready.
 * @type {Boolean}
 * @private
 */
ModelBase.prototype.isTrackStorageReady_ = false;

/**
 * Current track collection name.
 * @type {String}
 * @private
 */
TrackableModel.prototype.trackCollectionName_ = '';

//noinspection JSValidateJSDoc
/**
 * Track storage.
 * @type {mongodb.Collection}
 * @private
 */
TrackableModel.prototype.trackStorage_ = null;

/**
 * Handle all state changes and save new state to database.
 */
TrackableModel.prototype.changeTrackHandler_ = function () {

	var self = this,
		modelState = self.serialize(),
		execute = function () {

			self.trackStorage_.insert({
					modelName: self.name,
					timestamp: new Date().getTime(),
					state: modelState
				},
				{safe: true},
				function (error) {
					if (error) {
						throw error;
					}
				});
		};

	if (self.isTrackStorageReady_) {
		execute();
		return;
	}

	self.once(TrackableModel.events.trackStorageReady, execute);
};