/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var util = require('util'),
	mongodb = require('mongodb'),
	crypto = require('crypto'),
	uuid = require('node-uuid'),
	events = require('events'),
	Scheduler = require('../common/patterns/scheduler.js'),
	DatabaseProvider = require('../io/database-provider.js'),
	singleton = require('../common/patterns/singleton.js');

module.exports = singleton(SessionManager);

util.inherits(SessionManager, events.EventEmitter);

function SessionManager(config) {

	var self = this;

	events.EventEmitter.call(self);

	self.collectionName_ = config.$(SessionManager.parameters.collectionName);
	self.expiredTimeout_ = config.$(SessionManager.parameters.expiredTimeout);
	self.checkExpiredInterval_ = config.$(SessionManager.parameters.checkExpiredInterval);

	self.dbProvider_ = new DatabaseProvider(config);
	var dbProviderEvents = DatabaseProvider.events;
	self.dbProvider_.on(dbProviderEvents.error, function (error) {
		self.emit(SessionManager.events.error, error);
	});

	self.on(SessionManager.events.storageReady, function () {
		self.isStorageReady_ = true;
	});

	self.dbProvider_.getCollection(self.collectionName_, function (collection) {
		self.storage_ = collection;
		self.emit(SessionManager.events.storageReady);
	});

	self.checkExpiredScheduler_ = new Scheduler(0, self.checkExpiredInterval_, function () {
		self.checkExpired();
	});

	self.checkExpiredScheduler_.start();

}

/**
 * Enum of session manager events.
 * @enum {String}
 */
SessionManager.events = {
	storageReady: 'storageReady',
	error: 'error'
};

/**
 * Sessions related parameters.
 * @type {Object}
 */
SessionManager.parameters = {
	collectionName: 'sessions.collectionName',
	expiredTimeout: 'sessions.expiredTimeout',
	checkExpiredInterval: 'sessions.checkExpiredInterval'
};

/**
 * Is session storage ready.
 * @type {Boolean}
 */
SessionManager.prototype.isStorageReady_ = false;

/**
 * Name of session storage in database.
 * @type {String}
 * @private
 */
SessionManager.prototype.collectionName_ = '';

/**
 * Count of minutes before session will be destroyed.
 * @type {Number}
 * @private
 */
SessionManager.prototype.expiredTimeout_ = 0;

/**
 * Interval in milliseconds to collect garbage in session storage.
 * @type {Number}
 * @private
 */
SessionManager.prototype.checkExpiredInterval_ = 0;

/**
 * Database provider object.
 * @type {DatabaseProvider}
 * @private
 */
SessionManager.prototype.dbProvider_ = null;

//noinspection JSValidateJSDoc
/**
 * Session database collection.
 * @type {mongodb.Collection}
 * @private
 */
SessionManager.prototype.storage_ = null;

/**
 * Garbage collector scheduler.
 * @type {Scheduler}
 * @private
 */
SessionManager.prototype.checkExpiredScheduler_ = null;

/**
 * Create new session for user.
 * @param {Object} sessionObject New session object.
 * @param {function(Error,Object,String)} callback Session creation result handler.
 */
SessionManager.prototype.create = function (sessionObject, callback) {

	var self = this,
		now = new Date().getTime(),
		expiredDate = now + self.expiredTimeout_,
		execute = function () {
			sessionObject.expired = expiredDate;
			sessionObject.sid = self.createSessionId();
			self.storage_.insert(sessionObject, {safe: true}, function (error, objects) {

				if (error) {
					self.emit(SessionManager.events.error, error);
					return;
				}

				var object = objects[0];

				callback(error, object, !error ? object.sid : null);

			});
		};

	if (self.isStorageReady_) {
		execute();
	} else {
		self.once(SessionManager.events.storageReady, execute);
	}
};

/**
 * Remove session of specified user.
 * @param {String} sessionId ID of session which need to remove..
 * @param {function(Error)} callback Operation result handler.
 */
SessionManager.prototype.remove = function (sessionId, callback) {

	var self = this,
		execute = function () {

			self.storage_.remove({sid: sessionId}, {safe: true}, function (error) {

				if (error) {
					self.emit(SessionManager.events.error, error);
				}

				callback(error);

			});
		};

	if (self.isStorageReady_) {
		execute();
	} else {
		self.once(SessionManager.events.storageReady, execute);
	}
};

/**
 * Get session object for specified session ID.
 * @param {String} sessionId Session ID to get object.
 * @param {function(Error,Object)} callback Session object handler.
 */
SessionManager.prototype.get = function (sessionId, callback) {

	var self = this,
		execute = function () {

			self.storage_.findOne({sid: sessionId}, function (error, object) {

				if (error) {
					self.emit(SessionManager.events.error, error);
				}

				callback(error, object);

			});
		};

	if (self.isStorageReady_) {
		execute();
	} else {
		self.once(SessionManager.events.storageReady, execute);
	}
};

/**
 * Set new session object for specified session ID.
 * @param {String} sessionId Session ID to set object.
 * @param {Object} setParameters Parameters to set.
 * @param {function(Error,Object)} callback
 */
SessionManager.prototype.set = function (sessionId, setParameters, callback) {

	var self = this,
		now = new Date().getTime(),
		expiredDate = now + self.expiredTimeout_,
		execute = function () {

			setParameters.expired = expiredDate;

			self.storage_.findAndModify({sid: sessionId}, [
				['_id', 'asc']
			], {$set: setParameters}, {new: true, upsert: false},
				function (error, object) {

					if (error) {
						self.emit(SessionManager.events.error, error);
					}

					callback(error, object);

				});

		};

	if (self.isStorageReady_) {
		execute();
	} else {
		self.once(SessionManager.events.storageReady, execute);
	}
};

/**
 * Get or set object for specified session ID.
 * @param {String} sessionId Session ID to set or get object.
 * @param {function(Error,Object)} callback Operation result handler.
 * @param {Object} setParameters Object to set for specified session ID (Optional).
 */
SessionManager.prototype.$ = function (sessionId, callback, setParameters) {

	if (setParameters === undefined) {
		this.get(sessionId, callback);
		return;
	}

	this.set(sessionId, setParameters, callback);
};

/**
 * Remove expired sessions from storage.
 */
SessionManager.prototype.checkExpired = function () {

	var self = this,
		now = new Date(),
		execute = function () {
			self.storage_.remove({'expired': {$lte: now.getTime()}}, function (error) {
				if (error) {
					self.emit(SessionManager.events.error, error);
				}
			});
		};

	if (self.isStorageReady_) {
		execute();
	} else {
		self.once(SessionManager.events.storageReady, execute);
	}

};

/**
 * Stop removing expired sessions.
 */
SessionManager.prototype.stopCheckExpireInterval = function () {
	this.checkExpiredScheduler_.stop();
};

/**
 * Create new unique session ID.
 * @return {String}
 */
SessionManager.prototype.createSessionId = function () {

	var uniqueKey = uuid.v1(),
		hash = crypto.createHash('sha512');

	hash.update(uniqueKey, 'utf8');
	return hash.digest('hex');

};