/* 
 * neutrino
 *
 * Copyright (c) 2012 Denis Rechkunov and project contributors.
 *
 * neutrino's license follows:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, 
 * publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, 
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * This license applies to all parts of neutrino that are not externally
 * maintained libraries.
 */

var neutrino = require('../../index.js'),
	loggerConfig = {
		"levels": ""
	},
	Config = require('../../core/config.js'),
	LogicSet = require('../../cluster/worker/logic-set.js');
ViewBase = require('../../mvc/view/view-base.js');
DatabaseProvider = require('../../io/database-provider.js');
neutrino.configure({logger: loggerConfig});

var dbProvider = new DatabaseProvider(neutrino.currentConfig),
	logicSetEvents = LogicSet.events,
	viewEvents = ViewBase.events,
	masterConfig = new Config({
		"eventBus": {
			"serverPort": 8085,
			"sendMessageQueueInterval": 1
		},
		logger: loggerConfig
	}),
	master = new neutrino.createMaster(masterConfig);

exports['Get model'] = function (test) {

	var config = {
			"eventBus": {
				"serverPort": 8085
			},
			worker: {
				mvc: {
					modelsCollectionName: 'testModelsViewTest1',
					modelsFolder: "./tests/models"
				},
				port: 50889
			},
			logger: loggerConfig
		},
		worker = neutrino.createWorker(config),
		logicSet = worker.logicSet_;

	test.expect(5);
	logicSet.on(logicSetEvents.loaded, function () {

		var myRequestId = new Date().getTime(),
			mySessionId = new Date().getTime() + Math.random();

		logicSet.views_['Test'].on(viewEvents.showResponse, function (responseObject, sessionId, requestId) {

			test.deepEqual(responseObject.success.true);
			test.notEqual(responseObject.model.properties, undefined);
			test.deepEqual(responseObject.model.properties.test, 'testValue');
			test.deepEqual(sessionId, mySessionId);
			test.deepEqual(requestId, myRequestId);

			dbProvider.getCollection(config.worker.mvc.modelsCollectionName, function (collection) {
				collection.drop();
				test.done();
			});
		});
		logicSet.views_['Test'].getModel(mySessionId, myRequestId);
	});
};

exports['Set value and receive update'] = function (test) {

	var config = {
			"eventBus": {
				"serverPort": 8085
			},
			worker: {
				mvc: {
					modelsCollectionName: 'testModelsViewTest2',
					modelsFolder: "./tests/models"
				},
				port: 50890
			},
			logger: loggerConfig
		},
		worker = neutrino.createWorker(config),
		logicSet = worker.logicSet_;

	test.expect(8);
	logicSet.on(logicSetEvents.loaded, function () {

		var myRequestId = new Date().getTime(),
			mySessionId = new Date().getTime() + Math.random();

		logicSet.views_['Test'].on(viewEvents.updateValue, function (propertyName, oldValue, newValue, sessionId) {

			test.deepEqual(logicSet.models_['Test'].test.$(), 'newTestValue');
			test.deepEqual(propertyName, 'test');
			test.deepEqual(oldValue, 'testValue');
			test.deepEqual(newValue, 'newTestValue');
			test.deepEqual(sessionId, mySessionId);

			dbProvider.getCollection(config.worker.mvc.modelsCollectionName, function (collection) {
				collection.drop();
				test.done();
			});
		});
		logicSet.views_['Test'].on(viewEvents.showResponse, function (responseObject, sessionId, requestId) {
			test.deepEqual(responseObject.success, true);
			test.deepEqual(sessionId, mySessionId);
			test.deepEqual(requestId, myRequestId);
		});
		logicSet.views_['Test'].subscribe(mySessionId, myRequestId);
		logicSet.views_['Test'].setValue('test', 'newTestValue', mySessionId, myRequestId);
	});
};

exports['Invoke method and receive a result'] = function (test) {

	var config = {
			"eventBus": {
				"serverPort": 8085
			},
			worker: {
				mvc: {
					modelsCollectionName: 'testModelsViewTest3',
					modelsFolder: "./tests/models"
				},
				port: 50891
			},
			logger: loggerConfig
		},
		worker = neutrino.createWorker(config),
		logicSet = worker.logicSet_;

	test.expect(4);
	logicSet.on(logicSetEvents.loaded, function () {

		var myRequestId = new Date().getTime(),
			mySessionId = new Date().getTime() + Math.random();

		logicSet.views_['Test'].on(viewEvents.showResponse, function (responseObject, sessionId, requestId) {
			test.deepEqual(responseObject.success, true);
			test.deepEqual(sessionId, mySessionId);
			test.deepEqual(requestId, myRequestId);
			test.deepEqual(responseObject.result, 'returnedTestValue');

			dbProvider.getCollection(config.worker.mvc.modelsCollectionName, function (collection) {
				collection.drop();
				test.done();
			});
		});

		logicSet.views_['Test'].invoke('testMethod', ['returnedTestValue'], mySessionId, myRequestId);
	});
};

exports['Get private property error delivery'] = function (test) {

	var config = {
			"eventBus": {
				"serverPort": 8085
			},
			worker: {
				mvc: {
					modelsCollectionName: 'testModelsViewTest4',
					modelsFolder: "./tests/models"
				},
				port: 50892
			},
			logger: loggerConfig
		},
		worker = neutrino.createWorker(config),
		logicSet = worker.logicSet_;

	test.expect(4);
	logicSet.on(logicSetEvents.loaded, function () {

		var myRequestId = new Date().getTime(),
			mySessionId = new Date().getTime() + Math.random();

		logicSet.views_['Test'].on(viewEvents.showResponse, function (responseObject, sessionId, requestId) {
			test.deepEqual(requestId, myRequestId);
			test.deepEqual(sessionId, mySessionId);
			test.deepEqual(responseObject.success, false);
			test.deepEqual(responseObject.error.message, 'Can not invoke private methods');
			dbProvider.getCollection(config.worker.mvc.modelsCollectionName, function (collection) {
				collection.drop();
				test.done();
			});
		});

		logicSet.views_['Test'].invoke('privateTest_', ['returnedTestValue'], mySessionId, myRequestId);
	});

};